import argparse
import time

import torch

from diffusers import StableVideoDiffusionPipeline
from diffusers.utils import load_image
from onediff.quantization import QuantPipeline
from PIL import Image


parser = argparse.ArgumentParser()
parser.add_argument(
    "--model", type=str, default="stabilityai/stable-video-diffusion-img2vid-xt"
)
parser.add_argument("--variant", type=str, default="fp16")
parser.add_argument("--quantized_model", type=str, required=True)
parser.add_argument("--input_image", type=str, required=True)
parser.add_argument("--steps", type=int, default=25)
parser.add_argument("--height", type=int, default=576)
parser.add_argument("--width", type=int, default=1024)
parser.add_argument(
    "--conv_compute_density_threshold",
    type=int,
    default=900,
    help="The conv modules whose computational density is higher than the threshold will be quantized.",
)
parser.add_argument(
    "--linear_compute_density_threshold",
    type=int,
    default=500,
    help="The linear modules whose computational density is higher than the threshold will be quantized.",
)
parser.add_argument(
    "--conv_ssim_threshold",
    type=float,
    default=0,
    help="A similarity threshold that quantize convolution. The higher the threshold, the lower the accuracy loss caused by quantization.",
)
parser.add_argument(
    "--linear_ssim_threshold",
    type=float,
    default=0,
    help="A similarity threshold that quantize linear. The higher the threshold, the lower the accuracy loss caused by quantization.",
)
parser.add_argument(
    "--save_as_float",
    default=False,
    type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
)
parser.add_argument("--seed", type=int, default=1)
parser.add_argument("--cache_dir", type=str, default=None)
args = parser.parse_args()

pipe = QuantPipeline.from_pretrained(
    StableVideoDiffusionPipeline,
    args.model,
    torch_dtype=torch.float16,
    variant=args.variant,
    use_safetensors=True,
)
pipe.to("cuda")

input_image = load_image(args.input_image)
input_image.resize((args.width, args.height), Image.LANCZOS)
pipe_kwargs = dict(
    image=input_image,
    height=args.height,
    width=args.width,
    num_inference_steps=args.steps,
    decode_chunk_size=5,
    seed=args.seed,
)

start_time = time.time()

pipe.quantize(
    **pipe_kwargs,
    conv_compute_density_threshold=args.conv_compute_density_threshold,
    linear_compute_density_threshold=args.linear_compute_density_threshold,
    conv_ssim_threshold=args.conv_ssim_threshold,
    linear_ssim_threshold=args.linear_ssim_threshold,
    save_as_float=args.save_as_float,
    cache_dir=args.cache_dir,
)

pipe.save_quantized(args.quantized_model, safe_serialization=True)

end_time = time.time()

print(f"Quantize module time: {end_time - start_time}s")



MODEL = "maxin-cn/Latte-1"
CKPT = "t2v_v20240523.pt"
VARIANT = None
CUSTOM_PIPELINE = None
# SAMPLE_METHOD = "DDIM"
BETA_START = 0.0001
BETA_END = 0.02
BREA_SCHEDULE = "linear"
VARIANCE_TYPE = "learned_range"
STEPS = 50
SEED = 25
WARMUPS = 1
BATCH = 1
HEIGHT = 512
WIDTH = 512
VIDEO_LENGTH = 16
FPS = 8
GUIDANCE_SCALE = 7.5
ENABLE_TEMPORAL_ATTENTIONS = "true"
ENABLE_VAE_TEMPORAL_DECODER = "true"
OUTPUT_VIDEO = "output.mp4"

PROMPT = "An epic tornado attacking above aglowing city at night."

EXTRA_CALL_KWARGS = None
ATTENTION_FP16_SCORE_ACCUM_MAX_M = 0

COMPILER_CONFIG = None


import argparse
import importlib
import inspect
import json
import os
import random
import time

import imageio

import torch
from diffusers.models import AutoencoderKL, AutoencoderKLTemporalDecoder
from diffusers.schedulers import DDIMScheduler
from onediffx import compile_pipe
from PIL import Image, ImageDraw
from transformers import T5EncoderModel, T5Tokenizer


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", type=str, default=MODEL)
    parser.add_argument("--ckpt", type=str, default=CKPT)
    parser.add_argument("--prompt", type=str, default=PROMPT)
    parser.add_argument("--save_graph", action="store_true")
    parser.add_argument("--load_graph", action="store_true")
    parser.add_argument("--variant", type=str, default=VARIANT)
    parser.add_argument("--custom-pipeline", type=str, default=CUSTOM_PIPELINE)
    # parser.add_argument("--sample-method", type=str, default=SAMPLE_METHOD)
    parser.add_argument("--beta-start", type=float, default=BETA_START)
    parser.add_argument("--beta-end", type=float, default=BETA_END)
    parser.add_argument("--beta-schedule", type=str, default=BREA_SCHEDULE)
    parser.add_argument(
        "--enable_temporal_attentions",
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
        default=ENABLE_TEMPORAL_ATTENTIONS,
    )
    parser.add_argument(
        "--enable_vae_temporal_decoder",
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
        default=ENABLE_VAE_TEMPORAL_DECODER,
    )
    parser.add_argument("--guidance-scale", type=float, default=GUIDANCE_SCALE)
    parser.add_argument("--variance-type", type=str, default=VARIANCE_TYPE)
    parser.add_argument("--steps", type=int, default=STEPS)
    parser.add_argument("--seed", type=int, default=SEED)
    parser.add_argument("--warmups", type=int, default=WARMUPS)
    parser.add_argument("--batch", type=int, default=BATCH)
    parser.add_argument("--height", type=int, default=HEIGHT)
    parser.add_argument("--width", type=int, default=WIDTH)
    parser.add_argument("--video-length", type=int, default=VIDEO_LENGTH)
    parser.add_argument("--fps", type=int, default=FPS)
    parser.add_argument("--extra-call-kwargs", type=str, default=EXTRA_CALL_KWARGS)
    parser.add_argument("--output-video", type=str, default=OUTPUT_VIDEO)
    parser.add_argument(
        "--compiler",
        type=str,
        default="nexfort",
        choices=["none", "nexfort", "compile"],
    )
    parser.add_argument(
        "--compiler-config",
        type=str,
        default=COMPILER_CONFIG,
    )
    parser.add_argument(
        "--attention-fp16-score-accum-max-m",
        type=int,
        default=ATTENTION_FP16_SCORE_ACCUM_MAX_M,
    )
    return parser.parse_args()


class IterationProfiler:
    def __init__(self):
        self.begin = None
        self.end = None
        self.num_iterations = 0

    def get_iter_per_sec(self):
        if self.begin is None or self.end is None:
            return None
        self.end.synchronize()
        dur = self.begin.elapsed_time(self.end)
        return self.num_iterations / dur * 1000.0

    def callback_on_step_end(self, pipe, i, t, callback_kwargs={}):
        if self.begin is None:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.begin = event
        else:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.end = event
            self.num_iterations += 1
        return callback_kwargs


def main():
    args = parse_args()

    if os.path.exists(args.model):
        model_path = args.model
    else:
        from huggingface_hub import snapshot_download

        model_path = snapshot_download(repo_id=args.model)

    torch.set_grad_enabled(False)
    device = "cuda" if torch.cuda.is_available() else "cpu"

    from models.latte_t2v import LatteT2V
    from sample.pipeline_latte import LattePipeline

    transformer_model = LatteT2V.from_pretrained(
        model_path, subfolder="transformer", video_length=args.video_length
    ).to(device, dtype=torch.float16)

    if args.enable_vae_temporal_decoder:
        vae = AutoencoderKLTemporalDecoder.from_pretrained(
            args.model, subfolder="vae_temporal_decoder", torch_dtype=torch.float16
        ).to(device)
    else:
        vae = AutoencoderKL.from_pretrained(
            args.model, subfolder="vae", torch_dtype=torch.float16
        ).to(device)
    tokenizer = T5Tokenizer.from_pretrained(args.model, subfolder="tokenizer")
    text_encoder = T5EncoderModel.from_pretrained(
        args.model, subfolder="text_encoder", torch_dtype=torch.float16
    ).to(device)

    # set eval mode
    transformer_model.eval()
    vae.eval()
    text_encoder.eval()

    scheduler = DDIMScheduler.from_pretrained(
        model_path,
        subfolder="scheduler",
        beta_start=args.beta_start,
        beta_end=args.beta_end,
        beta_schedule=args.beta_schedule,
        variance_type=args.variance_type,
        clip_sample=False,
    )

    pipe = LattePipeline(
        vae=vae,
        text_encoder=text_encoder,
        tokenizer=tokenizer,
        scheduler=scheduler,
        transformer=transformer_model,
    ).to(device)

    if args.compiler == "none":
        pass
    elif args.compiler == "nexfort":
        print("Nexfort backend is now active...")
        if args.compiler_config is not None:
            # config with dict
            options = json.loads(args.compiler_config)
        else:
            # config with string
            options = '{"mode": "max-optimize:max-autotune:freezing:benchmark:low-precision",             \
                "memory_format": "channels_last", "options": {"inductor.optimize_linear_epilogue": false, \
                "triton.fuse_attention_allow_fp16_reduction": false}}'
        pipe = compile_pipe(
            pipe, backend="nexfort", options=options, fuse_qkv_projections=True
        )
    elif args.compiler == "compile":
        if hasattr(pipe, "unet"):
            pipe.unet = torch.compile(pipe.unet)
        if hasattr(pipe, "transformer"):
            pipe.transformer = torch.compile(pipe.transformer)
        if hasattr(pipe, "controlnet"):
            pipe.controlnet = torch.compile(pipe.controlnet)
        pipe.vae = torch.compile(pipe.vae)
    else:
        raise ValueError(f"Unknown compiler: {args.compiler}")

    def get_kwarg_inputs():
        kwarg_inputs = dict(
            prompt=args.prompt,
            video_length=args.video_length,
            height=args.height,
            width=args.width,
            num_inference_steps=args.steps,
            guidance_scale=args.guidance_scale,
            enable_temporal_attentions=args.enable_temporal_attentions,
            num_images_per_prompt=1,
            mask_feature=True,
            enable_vae_temporal_decoder=args.enable_vae_temporal_decoder,
            **(
                dict()
                if args.extra_call_kwargs is None
                else json.loads(args.extra_call_kwargs)
            ),
        )
        return kwarg_inputs

    if args.warmups > 0:
        print("=======================================")
        print("Begin warmup")
        begin = time.time()
        for _ in range(args.warmups):
            pipe(**get_kwarg_inputs()).video
        end = time.time()
        print("End warmup")
        print(f"Warmup time: {end - begin:.3f}s")

        print("=======================================")

    kwarg_inputs = get_kwarg_inputs()
    iter_profiler = IterationProfiler()
    if "callback_on_step_end" in inspect.signature(pipe).parameters:
        kwarg_inputs["callback_on_step_end"] = iter_profiler.callback_on_step_end
    elif "callback" in inspect.signature(pipe).parameters:
        kwarg_inputs["callback"] = iter_profiler.callback_on_step_end
    torch.manual_seed(args.seed)
    begin = time.time()
    videos = pipe(**kwarg_inputs).video
    end = time.time()

    print("=======================================")
    print(f"Inference time: {end - begin:.3f}s")
    iter_per_sec = iter_profiler.get_iter_per_sec()
    if iter_per_sec is not None:
        print(f"Iterations per second: {iter_per_sec:.3f}")
    cuda_mem_max_used = torch.cuda.max_memory_allocated() / (1024**3)
    cuda_mem_max_reserved = torch.cuda.max_memory_reserved() / (1024**3)
    print(f"Max used CUDA memory : {cuda_mem_max_used:.3f}GiB")
    print(f"Max reserved CUDA memory : {cuda_mem_max_reserved:.3f}GiB")
    print("=======================================")

    if args.output_video is not None:
        # export_to_video(output_frames[0], args.output_video, fps=args.fps)
        try:
            imageio.mimwrite(
                args.output_video, videos[0], fps=8, quality=9
            )  # highest quality is 10, lowest is 0
        except:
            print("Error when saving {}".format(args.prompt))

    else:
        print("Please set `--output-video` to save the output video")


if __name__ == "__main__":
    main()



""" image to image graph load ,a old example"""
import gc
import os
import shutil
import tempfile
import time
import unittest

import numpy as np
from PIL import Image
import oneflow as flow  # usort: skip
import oneflow as torch

from diffusers import EulerDiscreteScheduler, utils

from onediff import (
    OneFlowStableDiffusionImg2ImgPipeline as StableDiffusionImg2ImgPipeline,
)

from onediff.infer_compiler.backends.oneflow.utils.cost_util import cost_cnt


_MODEL_ID = "stabilityai/stable-diffusion-2"
_WITH_IMAGE_SAVE = True


def _reset_session():
    # Close session to avoid the buffer name duplicate error.
    flow.framework.session_context.TryCloseDefaultSession()
    time.sleep(5)
    flow.framework.session_context.NewDefaultSession(flow._oneflow_global_unique_env)


def _test_sd_graph_save_and_load(
    is_save, graph_save_path, sch_file_path, pipe_file_path
):
    if is_save:
        print("\n==> Try to run graph save...")
        _online_mode = False
        _pipe_from_file = False
    else:
        print("\n==> Try to run graph load...")
        _online_mode = True
        _pipe_from_file = True

    total_start_t = time.time()
    start_t = time.time()

    @cost_cnt
    def get_pipe():
        if _pipe_from_file:
            scheduler = EulerDiscreteScheduler.from_pretrained(
                sch_file_path, subfolder="scheduler"
            )
            sd_pipe = StableDiffusionImg2ImgPipeline.from_pretrained(
                pipe_file_path,
                scheduler=scheduler,
                revision="fp16",
                torch_dtype=torch.float16,
            )
        else:
            scheduler = EulerDiscreteScheduler.from_pretrained(
                _MODEL_ID, subfolder="scheduler"
            )
            sd_pipe = StableDiffusionImg2ImgPipeline.from_pretrained(
                _MODEL_ID,
                scheduler=scheduler,
                revision="fp16",
                torch_dtype=torch.float16,
            )

        return scheduler, sd_pipe

    sch, pipe = get_pipe()

    @cost_cnt
    def pipe_to_cuda():
        cu_pipe = pipe.to("cuda")
        return cu_pipe

    pipe = pipe_to_cuda()

    @cost_cnt
    def config_graph():
        pipe.set_graph_compile_cache_size(9)
        pipe.enable_graph_share_mem()

    config_graph()

    if not _online_mode:
        pipe.enable_save_graph()
    else:

        @cost_cnt
        def load_graph():
            assert os.path.exists(graph_save_path) and os.path.isdir(graph_save_path)
            pipe.load_graph(graph_save_path, compile_unet=True, compile_vae=True)

        load_graph()
    end_t = time.time()
    print("sd init time ", end_t - start_t, "s.")

    @cost_cnt
    def image_to_image(
        prompt, img, num_images_per_prompt=1, prefix="", with_graph=False
    ):
        cur_generator = torch.Generator("cuda").manual_seed(1024)
        strength = 0.8
        num_inference_steps = 100
        guidance_scale = 10
        eta = 0.0

        images = pipe(
            prompt,
            image=img,
            strength=strength,
            num_inference_steps=num_inference_steps,
            guidance_scale=guidance_scale,
            compile_unet=with_graph,
            compile_vae=with_graph,
            num_images_per_prompt=num_images_per_prompt,
            eta=eta,
            generator=cur_generator,
            output_type="np",
        ).images

        if _WITH_IMAGE_SAVE:
            for i, image in enumerate(images):
                pipe.numpy_to_pil(image)[0].save(
                    f"{prefix}{prompt}_{i}-with_graph_{str(with_graph)}.png"
                )

        return images

    prompt = "sea,beach,the waves crashed on the sand,blue sky whit white cloud"
    img = Image.new("RGB", (512, 512), "#1f80f0")

    no_g_images = image_to_image(
        prompt, img, prefix=f"is_save_{str(is_save)}-", with_graph=False
    )
    with_g_images = image_to_image(
        prompt, img, prefix=f"is_save_{str(is_save)}-", with_graph=True
    )
    assert len(no_g_images) == len(with_g_images)
    for img_idx in range(len(no_g_images)):
        print(
            "====> diff ", np.abs(no_g_images[img_idx] - with_g_images[img_idx]).mean()
        )
        assert np.abs(no_g_images[img_idx] - with_g_images[img_idx]).mean() < 1e-2

    total_end_t = time.time()
    print("st init and run time ", total_end_t - total_start_t, "s.")

    @cost_cnt
    def save_pipe_sch():
        pipe.save_pretrained(pipe_file_path)
        sch.save_pretrained(sch_file_path)

    @cost_cnt
    def save_graph():
        assert os.path.exists(graph_save_path) and os.path.isdir(graph_save_path)
        pipe.save_graph(graph_save_path)

    if not _online_mode:
        save_pipe_sch()
        save_graph()


class OneFlowPipeLineGraphSaveLoadTests(unittest.TestCase):
    def tearDown(self):
        # clean up the VRAM after each test
        super().tearDown()
        gc.collect()
        torch.cuda.empty_cache()

    def test_sd_graph_save_and_load(self):
        with tempfile.TemporaryDirectory() as f0:
            with tempfile.TemporaryDirectory() as f1:
                with tempfile.TemporaryDirectory() as f2:
                    _test_sd_graph_save_and_load(True, f0, f1, f2)
                    _reset_session()
                    _test_sd_graph_save_and_load(False, f0, f1, f2)


if __name__ == "__main__":
    unittest.main()



import functools

from onediff.utils.log_utils import logger

from ..hijack_ipadapter_plus.set_model_patch_replace import set_model_patch_replace_v2

from ..utils.booster_utils import is_using_oneflow_backend
from ._config import comfyui_instantid_hijacker, comfyui_instantid_pt

set_model_patch_replace_fn_pt = comfyui_instantid_pt.InstantID._set_model_patch_replace
apply_instantid = comfyui_instantid_pt.InstantID.ApplyInstantID.apply_instantid
InstantID = comfyui_instantid_pt.InstantID.InstantID
patch_attention = comfyui_instantid_pt.InstantID.InstantIDAttentionPatch.patch_attention


def cache_init(original_init):
    @functools.wraps(original_init)
    def new_init(self, instantid_model, *args, **kwargs):
        cache_key = args + tuple(kwargs.values())
        # import time
        # t0 = time.time()
        if cache_key in instantid_model:
            cached_instance = instantid_model[cache_key]
            self.__dict__ = cached_instance.__dict__
            logger.debug("Using cached InstantID instance")
        else:
            original_init(self, instantid_model, *args, **kwargs)
            instantid_model[cache_key] = self
            logger.debug("Caching new InstantID instance")

        # print(f'dur: {time.time() - t0}')

    return new_init


def cond_func(org_fn, model, *args, **kwargs):
    return is_using_oneflow_backend(model)


def apply_instantid_of(org_fn, self, *args, **kwargs):
    model = kwargs["model"]
    _org_init = InstantID.__init__
    InstantID.__init__ = cache_init(_org_init)
    output = org_fn(self, *args, **kwargs)
    InstantID.__init__ = _org_init
    return output


def apply_instantid_cond(org_fn, self, *args, **kwargs):
    return is_using_oneflow_backend(kwargs["model"])


comfyui_instantid_hijacker.register(
    set_model_patch_replace_fn_pt, set_model_patch_replace_v2, cond_func
)

comfyui_instantid_hijacker.register(
    apply_instantid, apply_instantid_of, apply_instantid_cond
)

comfyui_instantid_hijacker.register(
    patch_attention, apply_instantid_of, apply_instantid_cond
)



"""
code from https://github.com/comfyanonymous/ComfyUI/blob/4103f7fad5be7e22ed61843166b72b7c41671d75/comfy/ldm/modules/attention.py#L450-L490
"""
from abc import abstractmethod
from typing import Any, Optional

import comfy
import oneflow as torch
import oneflow.nn as nn
from einops import rearrange, repeat
from onediff.infer_compiler.backends.oneflow.transform import proxy_class, transform_mgr

onediff_comfy = transform_mgr.transform_package("comfy")

ops = onediff_comfy.ops.disable_weight_init
timestep_embedding = onediff_comfy.ldm.modules.diffusionmodules.util.timestep_embedding


def exists(val):
    return val is not None


def uniq(arr):
    return {el: True for el in arr}.keys()


def default(val, d):
    if exists(val):
        return val
    return d


class SpatialTransformer(proxy_class(comfy.ldm.modules.attention.SpatialTransformer)):
    """
    Transformer block for image-like data.
    First, project the input (aka embedding)
    and reshape to b, t, d.
    Then apply standard transformer action.
    Finally, reshape to image
    NEW: use_linear for more efficiency instead of the 1x1 convs
    """

    def __init__(
        self,
        in_channels,
        n_heads,
        d_head,
        depth=1,
        dropout=0.0,
        context_dim=None,
        disable_self_attn=False,
        use_linear=False,
        use_checkpoint=True,
        dtype=None,
        device=None,
        operations=comfy.ops,
    ):
        super().__init__()
        if exists(context_dim) and not isinstance(context_dim, list):
            context_dim = [context_dim] * depth
        self.in_channels = in_channels
        inner_dim = n_heads * d_head
        self.norm = Normalize(in_channels, dtype=dtype, device=device)
        if not use_linear:
            self.proj_in = operations.Conv2d(
                in_channels,
                inner_dim,
                kernel_size=1,
                stride=1,
                padding=0,
                dtype=dtype,
                device=device,
            )
        else:
            self.proj_in = operations.Linear(
                in_channels, inner_dim, dtype=dtype, device=device
            )

        self.transformer_blocks = nn.ModuleList(
            [
                BasicTransformerBlock(
                    inner_dim,
                    n_heads,
                    d_head,
                    dropout=dropout,
                    context_dim=context_dim[d],
                    disable_self_attn=disable_self_attn,
                    checkpoint=use_checkpoint,
                    dtype=dtype,
                    device=device,
                    operations=operations,
                )
                for d in range(depth)
            ]
        )
        if not use_linear:
            self.proj_out = operations.Conv2d(
                inner_dim,
                in_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                dtype=dtype,
                device=device,
            )
        else:
            self.proj_out = operations.Linear(
                in_channels, inner_dim, dtype=dtype, device=device
            )
        self.use_linear = use_linear

    def forward(self, x, context=None, transformer_options={}):
        # note: if no context is given, cross-attention defaults to self-attention
        if not isinstance(context, list):
            context = [context] * len(self.transformer_blocks)
        b, c, h, w = x.shape
        x_in = x
        x = self.norm(x)
        if not self.use_linear:
            x = self.proj_in(x)
        # NOTE: rearrange in ComfyUI is replaced with reshape and use -1 to enable for
        # dynamic shape inference (multi resolution compilation)
        x = x.flatten(2, 3).permute(0, 2, 1)
        # x = x.reshape(b, c, -1).permute(0, 2, 1)
        # x = rearrange(x, 'b c h w -> b (h w) c').contiguous()
        if self.use_linear:
            x = self.proj_in(x)
        for i, block in enumerate(self.transformer_blocks):
            transformer_options["block_index"] = i
            x = block(x, context=context[i], transformer_options=transformer_options)
        if self.use_linear:
            x = self.proj_out(x)
        # NOTE: rearrange in ComfyUI is replaced with permute
        x = x.permute(0, 2, 1).reshape_as(x_in)
        # x = rearrange(x, 'b (h w) c -> b c h w', h=h, w=w).contiguous()
        if not self.use_linear:
            x = self.proj_out(x)
        return x + x_in


class CrossAttention(proxy_class(comfy.ldm.modules.attention.CrossAttention)):
    def __init__(
        self,
        query_dim,
        context_dim=None,
        heads=8,
        dim_head=64,
        dropout=0.0,
        dtype=None,
        device=None,
        operations=comfy.ops,
    ):
        super().__init__()
        inner_dim = dim_head * heads
        context_dim = default(context_dim, query_dim)

        self.heads = heads
        self.dim_head = dim_head

        self.to_q = operations.Linear(
            query_dim, inner_dim, bias=False, dtype=dtype, device=device
        )
        self.to_k = operations.Linear(
            context_dim, inner_dim, bias=False, dtype=dtype, device=device
        )
        self.to_v = operations.Linear(
            context_dim, inner_dim, bias=False, dtype=dtype, device=device
        )

        self.to_out = nn.Sequential(
            operations.Linear(inner_dim, query_dim, dtype=dtype, device=device),
            nn.Dropout(dropout),
        )
        self.attention_op: Optional[Any] = None

    # None of attention methods from the original codes is utilized
    # https://github.com/comfyanonymous/ComfyUI/blob/777f6b15225197898a5f49742682a2be859072d7/comfy/ldm/modules/attention.py#L333-L351
    def forward(self, x, context=None, value=None, mask=None):
        if (
            not exists(context)
            and not exists(value)
            and hasattr(self, "to_qkv")
            and self.to_qkv is not None
        ):
            qkv = self.to_qkv(x)
            out = torch._C.fused_multi_head_attention_inference_v2(
                query=qkv,
                query_head_size=self.dim_head,
                query_layout="BM(H3K)",
                output_layout="BM(HK)",
                scale=self.scale,
                causal=False,
            )
            return self.to_out(out)

        q = self.to_q(x)
        context = default(context, x)
        k = self.to_k(context)
        if value is not None:
            v = self.to_v(value)
            del value
        else:
            v = self.to_v(context)

        b, _, _ = q.shape

        head_dim = self.dim_head
        out = torch._C.fused_multi_head_attention_inference_v2(
            query=q,
            query_layout="BM(HK)",
            query_head_size=head_dim,
            key=k,
            key_layout="BM(HK)",
            value=v,
            value_layout="BM(HK)",
            output_layout="BM(HK)",
            causal=False,
        )

        if exists(mask):
            raise NotImplementedError

        return self.to_out(out)


class SpatialVideoTransformer(
    proxy_class(comfy.ldm.modules.attention.SpatialVideoTransformer)
):
    def __init__(
        self,
        in_channels,
        n_heads,
        d_head,
        depth=1,
        dropout=0.0,
        use_linear=False,
        context_dim=None,
        use_spatial_context=False,
        timesteps=None,
        merge_strategy: str = "fixed",
        merge_factor: float = 0.5,
        time_context_dim=None,
        ff_in=False,
        checkpoint=False,
        time_depth=1,
        disable_self_attn=False,
        disable_temporal_crossattention=False,
        max_time_embed_period: int = 10000,
        dtype=None,
        device=None,
        operations=ops,
    ):
        super().__init__(
            in_channels,
            n_heads,
            d_head,
            depth=depth,
            dropout=dropout,
            use_checkpoint=checkpoint,
            context_dim=context_dim,
            use_linear=use_linear,
            disable_self_attn=disable_self_attn,
            dtype=dtype,
            device=device,
            operations=operations,
        )
        self.time_depth = time_depth
        self.depth = depth
        self.max_time_embed_period = max_time_embed_period

        time_mix_d_head = d_head
        n_time_mix_heads = n_heads

        time_mix_inner_dim = int(time_mix_d_head * n_time_mix_heads)

        inner_dim = n_heads * d_head
        if use_spatial_context:
            time_context_dim = context_dim

        self.time_stack = nn.ModuleList(
            [
                BasicTransformerBlock(
                    inner_dim,
                    n_time_mix_heads,
                    time_mix_d_head,
                    dropout=dropout,
                    context_dim=time_context_dim,
                    # timesteps=timesteps,
                    checkpoint=checkpoint,
                    ff_in=ff_in,
                    inner_dim=time_mix_inner_dim,
                    disable_self_attn=disable_self_attn,
                    disable_temporal_crossattention=disable_temporal_crossattention,
                    dtype=dtype,
                    device=device,
                    operations=operations,
                )
                for _ in range(self.depth)
            ]
        )

        assert len(self.time_stack) == len(self.transformer_blocks)

        self.use_spatial_context = use_spatial_context
        self.in_channels = in_channels

        time_embed_dim = self.in_channels * 4
        self.time_pos_embed = nn.Sequential(
            operations.Linear(
                self.in_channels, time_embed_dim, dtype=dtype, device=device
            ),
            nn.SiLU(),
            operations.Linear(
                time_embed_dim, self.in_channels, dtype=dtype, device=device
            ),
        )

        self.time_mixer = AlphaBlender(
            alpha=merge_factor, merge_strategy=merge_strategy
        )

    def forward(
        self,
        x: torch.Tensor,
        context: Optional[torch.Tensor] = None,
        time_context: Optional[torch.Tensor] = None,
        timesteps: Optional[int] = None,
        image_only_indicator: Optional[torch.Tensor] = None,
        transformer_options={},
    ) -> torch.Tensor:
        _, _, h, w = x.shape
        x_in = x
        spatial_context = None
        if exists(context):
            spatial_context = context
        if self.use_spatial_context:
            assert (
                context.ndim == 3
            ), f"n dims of spatial context should be 3 but are {context.ndim}"

            if time_context is None:
                time_context = context
            time_context_first_timestep = time_context[::timesteps]
            # time_context = repeat(
            #     time_context_first_timestep, "b ... -> (b n) ...", n=h * w
            # )
            # Rewrite for onediff SVD dynamic shape
            time_context = torch._C.broadcast_dim_like(
                time_context_first_timestep[None, :],
                x.flatten(2, 3),
                dim=0,
                like_dim=2,
            ).flatten(0, 1)

        elif time_context is not None and not self.use_spatial_context:
            time_context = repeat(time_context, "b ... -> (b n) ...", n=h * w)
            time_context = torch._C.broadcast_dim_like(
                time_context_first_timestep[None, :],
                x.flatten(2, 3),
                dim=0,
                like_dim=2,
            )
            if time_context.ndim == 2:
                time_context = rearrange(time_context, "b c -> b 1 c")
                # time_context = time_context.unsqueeze(1)

        x = self.norm(x)
        if not self.use_linear:
            x = self.proj_in(x)
        # x = rearrange(x, "b c h w -> b (h w) c")
        # Rewrite for onediff SVD dynamic shape
        x = x.permute(0, 2, 3, 1).flatten(1, 2)
        if self.use_linear:
            x = self.proj_in(x)

        num_frames = torch.arange(timesteps, device=x.device)
        num_frames = repeat(num_frames, "t -> b t", b=x.shape[0] // timesteps)
        # num_frames = rearrange(num_frames, "b t -> (b t)")
        # Rewrite for onediff SVD dynamic shape
        num_frames = num_frames.flatten()
        t_emb = timestep_embedding(
            num_frames,
            self.in_channels,
            repeat_only=False,
            max_period=self.max_time_embed_period,
        ).to(x.dtype)
        emb = self.time_pos_embed(t_emb)
        emb = emb[:, None, :]

        for it_, (block, mix_block) in enumerate(
            zip(self.transformer_blocks, self.time_stack)
        ):
            transformer_options["block_index"] = it_
            x = block(
                x,
                context=spatial_context,
                transformer_options=transformer_options,
            )

            x_mix = x
            x_mix = x_mix + emb

            B, S, C = x_mix.shape
            # x_mix = rearrange(x_mix, "(b t) s c -> (b s) t c", t=timesteps)
            # Rewrite for onediff SVD dynamic shape
            b = B // timesteps
            x_mix = x_mix.unflatten(0, shape=(b, -1)).permute(0, 2, 1, 3).flatten(0, 1)
            x_mix = mix_block(x_mix, context=time_context)  # TODO: transformer_options
            # x_mix = rearrange(
            #     x_mix, "(b s) t c -> (b t) s c", s=S, b=B // timesteps, c=C, t=timesteps
            # )
            # Rewrite for onediff SVD dynamic shape
            x_mix = x_mix.unflatten(0, shape=(b, -1)).permute(0, 2, 1, 3).flatten(0, 1)

            x = self.time_mixer(
                x_spatial=x, x_temporal=x_mix, image_only_indicator=image_only_indicator
            )

        if self.use_linear:
            x = self.proj_out(x)
        # x = rearrange(x, "b (h w) c -> b c h w", h=h, w=w)
        # Rewrite for onediff SVD dynamic shape
        x = x.reshape_as(x_in.permute(0, 2, 3, 1)).permute(0, 3, 1, 2).contiguous()
        if not self.use_linear:
            x = self.proj_out(x)
        out = x + x_in
        return out


def attention_pytorch_oneflow(q, k, v, heads, mask=None, attn_precision=None):
    b, _, dim_head = q.shape
    dim_head //= heads
    head_dim = dim_head
    out = torch._C.fused_multi_head_attention_inference_v2(
        query=q,
        query_layout="BM(HK)",
        query_head_size=head_dim,
        key=k,
        key_layout="BM(HK)",
        value=v,
        value_layout="BM(HK)",
        output_layout="BM(HK)",
        causal=False,
    )
    return out



from comfy.model_patcher import ModelPatcher
from onediff.utils import logger

from ..sd_hijack_utils import Hijacker
from .patch_management import create_patch_executor, PatchType
from .utils.booster_utils import is_using_oneflow_backend


def clone_oneflow(org_fn, self, *args, **kwargs):
    n = org_fn(self, *args, **kwargs)
    create_patch_executor(PatchType.UiNodeWithIndexPatch).copy_to(self, n)
    dc_patch_executor = create_patch_executor(PatchType.DCUNetExecutorPatch)
    if dc_patch_executor.check_patch(self):
        dc_patch_executor.copy_to(self, n)
    return n


def cond_func(org_fn, self):
    return is_using_oneflow_backend(self)


def unpatch_model_oneflow(org_fn, self, device_to=None, unpatch_weights=True):
    if unpatch_weights:
        logger.warning(
            f"{type(self.model.diffusion_model)} is quantized by onediff, so unpatching is skipped."
        )
    return


def unpatch_model_cond_func(org_fn, self, *args, **kwargs):
    if hasattr(self.model, "diffusion_model") and hasattr(
        self.model.diffusion_model, "_deployable_module_quantized"
    ):
        return self.model.diffusion_model._deployable_module_quantized
    return False


model_patch_hijacker = Hijacker()

model_patch_hijacker.register(ModelPatcher.clone, clone_oneflow, cond_func)
model_patch_hijacker.register(
    ModelPatcher.unpatch_model, unpatch_model_oneflow, unpatch_model_cond_func
)



from dataclasses import dataclass
from functools import singledispatchmethod
from typing import Optional

from comfy.model_patcher import ModelPatcher

from ..booster_interface import BoosterExecutor
from .utils.booster_utils import set_compiled_options
from .utils.deep_cache_speedup import deep_cache_speedup
from .utils.graph_path import generate_graph_path


@dataclass
class DeepcacheBoosterExecutor(BoosterExecutor):
    cache_interval: int = 3
    cache_layer_id: int = 0
    cache_block_id: int = 1
    start_step: int = 0
    end_step: int = 1000

    @singledispatchmethod
    def execute(self, model):
        print(
            "Warning: DeepcacheBoosterExecutor.apply is not implemented for model type:",
            type(model),
        )
        return model

    @execute.register(ModelPatcher)
    def _(
        self, model, ckpt_name: Optional[str] = None, use_graph=True, **kwargs
    ) -> ModelPatcher:
        model = deep_cache_speedup(
            model=model,
            use_graph=use_graph,
            cache_interval=self.cache_interval,
            cache_layer_id=self.cache_layer_id,
            cache_block_id=self.cache_block_id,
            start_step=self.start_step,
            end_step=self.end_step,
            use_oneflow_deepcache_speedup_modelpatcher=False,
        )[0]
        if ckpt_name:
            graph_file = generate_graph_path(
                ckpt_name, model.fast_deep_cache_unet._torch_module
            )
            set_compiled_options(model.fast_deep_cache_unet, graph_file)
            graph_file = generate_graph_path(
                ckpt_name, model.deep_cache_unet._torch_module
            )
            set_compiled_options(model.deep_cache_unet, graph_file)

        return model



from nodes import ControlNetApply, ControlNetApplyAdvanced

from ..sd_hijack_utils import Hijacker
from .onediff_controlnet import OneDiffControlLora


def apply_controlnet_base(orig_func, self, conditioning, control_net, image, strength):
    if strength == 0:
        return (conditioning,)

    c = []
    control_hint = image.movedim(-1, 1)
    for t in conditioning:
        n = [t[0], t[1].copy()]
        if not hasattr(self, "_c_net"):
            self._c_net = control_net.copy()

        c_net = self._c_net.set_cond_hint(control_hint, strength)

        if "control" in t[1]:
            c_net.set_previous_controlnet(t[1]["control"])
        n[1]["control"] = c_net
        n[1]["control_apply_to_uncond"] = True
        c.append(n)
    return (c,)


def apply_controlnet_cond_func_base(
    orig_func, self, conditioning, control_net, image, strength
):
    return isinstance(control_net, OneDiffControlLora)


def apply_controlnet_advanced(
    orig_func,
    self,
    positive,
    negative,
    control_net,
    image,
    strength,
    start_percent,
    end_percent,
):
    if strength == 0:
        return (positive, negative)
    control_hint = image.movedim(-1, 1)
    cnets = {}

    out = []
    for conditioning in [positive, negative]:
        c = []
        for t in conditioning:
            d = t[1].copy()

            prev_cnet = d.get("control", None)
            if prev_cnet in cnets:
                c_net = cnets[prev_cnet]
            else:
                if not hasattr(self, "_c_net"):
                    print("creating new cnet")
                    self._c_net = control_net.copy()
                c_net = self._c_net.set_cond_hint(
                    control_hint, strength, (start_percent, end_percent)
                )
                c_net.set_previous_controlnet(prev_cnet)
                cnets[prev_cnet] = c_net

            d["control"] = c_net
            d["control_apply_to_uncond"] = False
            n = [t[0], d]
            c.append(n)
        out.append(c)
    return (out[0], out[1])


def apply_controlnet_cond_func_advanced(
    orig_func,
    self,
    positive,
    negative,
    control_net,
    image,
    strength,
    start_percent,
    end_percent,
):
    return isinstance(control_net, OneDiffControlLora)


nodes_hijacker = Hijacker()
nodes_hijacker.register(
    orig_func=ControlNetApply.apply_controlnet,
    sub_func=apply_controlnet_base,
    cond_func=apply_controlnet_cond_func_base,
)
nodes_hijacker.register(
    orig_func=ControlNetApplyAdvanced.apply_controlnet,
    sub_func=apply_controlnet_advanced,
    cond_func=apply_controlnet_cond_func_advanced,
)



from comfy.model_base import BaseModel
from comfy.model_patcher import ModelPatcher
from onediff.infer_compiler.backends.nexfort.deployable_module import (
    NexfortDeployableModule as DeployableModule,
)


def clear_deployable_module_cache_and_unbind(*args, **kwargs):
    raise RuntimeError(f"TODO")


def is_using_nexfort_backend(module):
    if isinstance(module, ModelPatcher):
        if hasattr(module.model, "diffusion_model"):
            diff_model = module.model.diffusion_model
            return isinstance(diff_model, DeployableModule)
    if isinstance(module, BaseModel):
        if hasattr(module, "diffusion_model"):
            return isinstance(module.diffusion_model, DeployableModule)
    return False



from typing import Union

import oneflow

import torch
from comfy import model_management
from comfy.model_base import BaseModel, SVD_img2vid
from comfy.model_patcher import ModelPatcher

from onediff.infer_compiler.backends.oneflow import (
    OneflowDeployableModule as DeployableModule,
)
from onediff.utils import set_boolean_env_var
from onediff.utils.import_utils import is_oneflow_available

from ..patch_management import create_patch_executor, PatchType


def set_compiled_options(module: DeployableModule, graph_file="unet"):
    assert isinstance(module, DeployableModule)
    compile_options = module._deployable_module_options
    compile_options.graph_file = graph_file
    compile_options.graph_file_device = model_management.get_torch_device()


def is_fp16_model(model):
    """
    Check if the model is using FP16 (Half Precision).

    Parameters:
        model (torch.nn.Module): The PyTorch model.

    Returns:
        bool: True if the model is using FP16, False otherwise.
    """
    for param in model.parameters():
        if param.dtype == torch.float16:
            return True
    return False


def get_model_type(model):
    """
    Get the types of the parameters in the model.

    Parameters:
        model (torch.nn.Module): The PyTorch model.

    Returns:
        set: A set containing the types of the parameters in the model.
    """
    type_set = set()
    for param in model.parameters():
        type_set.add(param.dtype)
    return type_set


def set_environment_for_svd_img2vid(model: ModelPatcher):
    if isinstance(model, ModelPatcher) and isinstance(model.model, SVD_img2vid):
        # TODO(fengwen) check it affect performance
        # To avoid overflow issues while maintaining performance,
        # refer to: https://github.com/siliconflow/onediff/blob/09a94df1c1a9c93ec8681e79d24bcb39ff6f227b/examples/image_to_video.py#L112
        set_boolean_env_var(
            "ONEFLOW_ATTENTION_ALLOW_HALF_PRECISION_SCORE_ACCUMULATION_MAX_M", False
        )


def is_using_oneflow_backend(module):
    # First, check if oneflow is available and CUDA is enabled
    if is_oneflow_available() and not oneflow.cuda.is_available():
        print("OneFlow CUDA support is not available")
        return False

    # Check if the module
    if isinstance(module, oneflow.nn.Module):
        return True

    dc_patch_executor = create_patch_executor(PatchType.DCUNetExecutorPatch)
    if isinstance(module, ModelPatcher):
        deep_cache_module = dc_patch_executor.get_patch(module)
        if deep_cache_module[0] and isinstance(deep_cache_module[0], DeployableModule):
            return True
        if hasattr(module.model, "diffusion_model"):
            diff_model = module.model.diffusion_model
            return isinstance(diff_model, DeployableModule)
        else:
            return False

    if isinstance(module, BaseModel):
        if dc_patch_executor.is_use_deep_cache_unet(module):
            return True
        if hasattr(module, "diffusion_model"):
            return isinstance(module.diffusion_model, DeployableModule)
        else:
            return False

    if isinstance(module, DeployableModule):
        return True

    if hasattr(module, "parameters"):
        for param in module.parameters():
            if isinstance(param, oneflow.Tensor):
                return True

    warn_msg = (
        f"OneFlow backend is not detected for the module, the module is {type(module)}"
    )
    print(warn_msg)
    # If none of the above conditions are met, it's not using OneFlow backend
    return False


def clear_deployable_module_cache_and_unbind(
    module: Union[ModelPatcher, DeployableModule]
):
    if isinstance(module, ModelPatcher):
        dcu_patch = create_patch_executor(PatchType.DCUNetExecutorPatch)
        if dcu_patch.check_patch(module):
            for sub_module in dcu_patch.get_patch(module):
                sub_module._clear_old_graph()

        diff_model = module.model.diffusion_model
        if isinstance(diff_model, DeployableModule):
            diff_model._clear_old_graph()
        create_patch_executor(PatchType.CachedCrossAttentionPatch).clear_patch(
            diff_model
        )
        create_patch_executor(PatchType.UNetExtraInputOptions).clear_patch(diff_model)
    elif isinstance(module, DeployableModule):
        diff_model = module
        diff_model._clear_old_graph()
        create_patch_executor(PatchType.CachedCrossAttentionPatch).clear_patch(
            diff_model
        )
        create_patch_executor(PatchType.CrossAttentionForwardMasksPatch).clear_patch(
            diff_model
        )
    else:
        raise RuntimeError(f"Unexpected module type: {type(module)}.")



"""
This example shows how to reuse the compiled components of a pipeline to create new pipelines.

Usage:
    $ python reuse_compiled_pipeline_components.py --model_id <model_id>
"""
import argparse
from io import BytesIO

import PIL
import requests
import torch
from diffusers import (
    StableDiffusionImg2ImgPipeline,
    StableDiffusionInpaintPipeline,
    StableDiffusionPipeline,
)
from onediffx import compile_pipe


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model_id",
        type=str,
        default="runwayml/stable-diffusion-v1-5",
    )
    return parser.parse_args()


args = get_args()


def initialize_pipelines(model_id, model_params):
    pipeline = StableDiffusionPipeline.from_pretrained(model_id, **model_params)

    pipeline.to("cuda")
    pipeline = compile_pipe(pipeline)

    # Reuse the components of the pipeline to create new pipelines
    # Reference: https://huggingface.co/docs/diffusers/main/en/api/diffusion_pipeline#diffusers.DiffusionPipeline.components
    img2img_pipe = StableDiffusionImg2ImgPipeline(**pipeline.components)
    inpaint_pipe = StableDiffusionInpaintPipeline(**pipeline.components)
    return pipeline, img2img_pipe, inpaint_pipe


def download_image(url, resize=None):
    response = requests.get(url)
    image = PIL.Image.open(BytesIO(response.content)).convert("RGB")
    if resize:
        image = image.resize(resize)
    return image


def inference_text2img(pipe):
    prompt = "A fantasy landscape, trending on artstation"
    result_image = pipe(prompt=prompt).images[0]
    result_image.save("inference_text2img.png")


def inference_img2img(pipe):
    img_url = "https://raw.githubusercontent.com/CompVis/stable-diffusion/main/assets/stable-samples/img2img/sketch-mountains-input.jpg"
    init_image = download_image(img_url, resize=(768, 512))
    prompt = "A fantasy landscape, trending on artstation"
    result_image = pipe(
        prompt=prompt, image=init_image, strength=0.75, guidance_scale=7.5
    ).images[0]
    result_image.save("inference_img2img.png")


def inference_inpaint(pipe):
    img_url = "https://raw.githubusercontent.com/CompVis/latent-diffusion/main/data/inpainting_examples/overture-creations-5sI6fQgYIuo.png"
    mask_url = "https://raw.githubusercontent.com/CompVis/latent-diffusion/main/data/inpainting_examples/overture-creations-5sI6fQgYIuo_mask.png"

    init_image = download_image(img_url, resize=(512, 512))
    mask_image = download_image(mask_url, resize=(512, 512))

    prompt = "Face of a yellow cat, high resolution, sitting on a park bench"
    result_image = pipe(prompt=prompt, image=init_image, mask_image=mask_image).images[
        0
    ]
    result_image.save("inference_inpaint.png")


def main(model_id, model_params):
    text2img, img2img, inpaint = initialize_pipelines(model_id, model_params)

    inference_text2img(text2img)
    inference_img2img(img2img)
    inference_inpaint(inpaint)


if __name__ == "__main__":
    model_id = args.model_id
    model_params = {"torch_dtype": torch.float16}
    main(model_id, model_params)



"""Hijack utils for stable-diffusion."""
import importlib
import inspect
from collections import deque
from types import FunctionType
from typing import Callable, List, Union

__all__ = ["Hijacker", "hijack_func"]


def get_func_full_name(func: FunctionType):
    """Get the full name of a function."""
    module = inspect.getmodule(func)
    if module is None:
        raise ValueError(f"Cannot get module of function {func}")
    return f"{module.__name__}.{func.__qualname__}"


class CondFunc:
    """A function that conditionally calls another function.

    Copied from: https://github.com/AUTOMATIC1111/stable-diffusion-webui/blob/master/modules/sd_hijack_utils.py
    """

    # Dictionary to store hijacked methods and their corresponding CondFunc instances
    hijacked_registry = {}

    def __new__(
        cls,
        orig_func: Union[str, Callable],
        sub_funcs: List[FunctionType],
        cond_funcs: List[FunctionType],
    ):
        # self: CondFunc instance
        self = super(CondFunc, cls).__new__(cls)
        if isinstance(orig_func, FunctionType):
            orig_func = get_func_full_name(orig_func)

        assert isinstance(orig_func, str)

        func_path = orig_func.split(".")
        for i in range(len(func_path) - 1, -1, -1):
            try:
                resolved_obj = importlib.import_module(".".join(func_path[:i]))
                break
            except ImportError:
                pass

        if resolved_obj is None:
            raise ImportError(f"Could not resolve module for {func_path}")

        for attr_name in func_path[i:-1]:
            resolved_obj = getattr(resolved_obj, attr_name)

        orig_func = getattr(resolved_obj, func_path[-1])

        def hijacked_method(*args, **kwargs):
            return self(*args, **kwargs)

        setattr(
            resolved_obj,
            func_path[-1],
            hijacked_method,
        )

        def unhijack_func():
            setattr(resolved_obj, func_path[-1], orig_func)
            del cls.hijacked_registry[hijacked_method]

        self.__init__(orig_func, sub_funcs, cond_funcs)
        cls.hijacked_registry[hijacked_method] = self
        return (hijacked_method, unhijack_func)

    @staticmethod
    def is_hijacked_method(func: Callable):
        return func in CondFunc.hijacked_registry

    @staticmethod
    def get_hijacked_instance(func: Callable):
        return CondFunc.hijacked_registry.get(func)

    def __init__(
        self,
        orig_func: Callable,
        sub_funcs: List[FunctionType],
        cond_funcs: List[FunctionType],
    ):
        self._orig_func = orig_func
        self._sub_funcs = deque(sub_funcs)
        self._cond_funcs = deque(cond_funcs)

    def __call__(self, *args, **kwargs):
        for cond_func, sub_func in zip(self._cond_funcs, self._sub_funcs):
            if not cond_func or cond_func(self._orig_func, *args, **kwargs):
                return sub_func(self._orig_func, *args, **kwargs)
        else:
            return self._orig_func(*args, **kwargs)

    def add_condition(self, sub_func: FunctionType, cond_func: FunctionType, last=True):
        """Pairs are returned in LIFO order if last is true or FIFO order if false."""
        instance: CondFunc = self
        if last:
            instance._sub_funcs.append(sub_func)
            instance._cond_funcs.append(cond_func)
        else:
            instance._sub_funcs.appendleft(sub_func)
            instance._cond_funcs.appendleft(cond_func)


def ensure_list(obj: Union[FunctionType, List[FunctionType]]) -> List[FunctionType]:
    if not isinstance(obj, list):
        return [obj]
    return obj


def hijack_func(
    orig_func: Union[str, Callable],
    sub_func: Callable,
    cond_func: Callable,
    *,
    last=True,
):
    """
    Hijacks a function with another function.

    Returns:
        A tuple of (hijacked_func, unhijack_func)

    Examples:
        >>> def foo(*args, **kwargs):
        >>>     # orig_func
        >>>     print('foo')
        >>> def bar(orig_func, *args, **kwargs):
        >>>     # sub_func
        >>>     print('bar')
        >>> def cond_func(orig_func, *args, **kwargs):
        >>>     # cond_func
        >>>     return True
        >>> hijack_func(foo, bar, cond_func)
        >>> foo()
        bar
    """
    if CondFunc.is_hijacked_method(orig_func):
        ins = CondFunc.get_hijacked_instance(orig_func)
        ins.add_condition(sub_func, cond_func, last=last)
        return orig_func, lambda: None

    if isinstance(orig_func, FunctionType):
        orig_func = get_func_full_name(orig_func)
    return CondFunc(orig_func, ensure_list(sub_func), ensure_list(cond_func))


class Hijacker:
    """A class that hijacks a list of functions.

    Example:
        >>> hijacker = Hijacker([
        >>>     ("module.orig_func1", sub_func1, cond_func1),
        >>>     ("module.orig_func2", sub_func2, cond_func2),
        >>>     # Add more tuples as needed
        >>> ])
    """

    def __init__(self, funcs_list=[]):
        if funcs_list and isinstance(funcs_list, List):
            self.funcs_list = funcs_list.copy()
        else:
            self.funcs_list = []
        self.unhijack_funcs = []

    def hijack(self, last=True):
        self.unhijack()
        for orig_func, sub_func, cond_func in self.funcs_list:
            _, unhijack_func = hijack_func(orig_func, sub_func, cond_func, last=last)
            self.unhijack_funcs.append(unhijack_func)

    def unhijack(self):
        if len(self.unhijack_funcs) > 0:
            for unhijack_func in self.unhijack_funcs:
                unhijack_func()
            self.unhijack_funcs = []

    def extend_unhijack(self, unhijack_func):
        self.unhijack_funcs.append(unhijack_func)

    def register(
        self, orig_func: FunctionType, sub_func: Callable, cond_func: Callable
    ):
        self.funcs_list.append((orig_func, sub_func, cond_func))


if __name__ == "__main__":

    def orig_func(*args, **kwargs):
        print("Original function")
        return "orig_func"

    def sub_func_0(orig_func, *args, **kwargs):
        print(f"Called sub_func_0")
        return "sub_func_0"

    def sub_func_1(orig_func, *args, **kwargs):
        print(f"Called sub_func_1")
        return "sub_func_1"

    cond_0 = True

    def cond_func_0(orig_func, *args, **kwargs):
        return cond_0

    def cond_func_1(orig_func, *args, **kwargs):
        return True

    hijack_func(orig_func, sub_func_0, cond_func_0)
    assert orig_func() == "sub_func_0"  # Output: Called sub_func_0

    hijack_func(orig_func, sub_func_1, cond_func_1)
    cond_0 = False
    assert orig_func() == "sub_func_1"  # Output: Called sub_func_1
    cond_0 = True
    assert orig_func() == "sub_func_0"  # Called sub_func_0
    hijack_func(orig_func, sub_func_1, cond_func_1, last=False)
    cond_0 = True
    assert orig_func() == "sub_func_1"  # Called sub_func_1

    class Case1:
        def clone(self):
            print(f"{type(self)}.clone")

    def cond_func(org_fn, self):
        return True

    def custom_clone(org_fn, self):
        print(f"custom_clone")
        return "custom_clone"

    hijack_func(Case1.clone, custom_clone, cond_func)

    def custom_clone_1(org_fn, self):
        print(f"custom_clone_1")
        return "custom_clone_1"

    assert Case1().clone() == "custom_clone"
    hijack_func(Case1.clone, custom_clone_1, cond_func, last=False)

    assert Case1().clone() == "custom_clone_1"



import oneflow as flow  # usort: skip
from compile.oneflow.mock.common import timestep_embedding
from ldm.modules.diffusionmodules.openaimodel import UNetModel
from modules import devices

from onediff.infer_compiler.backends.oneflow.transform import proxy_class

cond_cast_unet = getattr(devices, "cond_cast_unet", lambda x: x)


# Due to the tracing mechanism in OneFlow, it's crucial to ensure that
# the same conditional branches are taken during the first run as in subsequent runs.
# Therefore, certain "optimizations" have been modified.
def aligned_adding(base, x, require_channel_alignment):
    if isinstance(x, float):
        # remove `if x == 0.0: return base` here
        return base + x

    if require_channel_alignment:
        zeros = flow.zeros_like(base)
        zeros[:, : x.shape[1], ...] = x
        x = zeros

    # resize to sample resolution
    base_h, base_w = base.shape[-2:]
    xh, xw = x.shape[-2:]

    if xh > 1 or xw > 1 and (base_h != xh or base_w != xw):
        # logger.info('[Warning] ControlNet finds unexpected mis-alignment in tensor shape.')
        x = flow.nn.functional.interpolate(x, size=(base_h, base_w), mode="nearest")
    return base + x


def cat(tensors, *args, **kwargs):
    if len(tensors) == 2:
        a, b = tensors
        a = flow.nn.functional.interpolate_like(a, like=b, mode="nearest")
        tensors = (a, b)
    return flow.cat(tensors, *args, **kwargs)


class OneFlowOnediffControlNetModel(proxy_class(UNetModel)):
    def forward(
        self,
        x,
        timesteps,
        context,
        y,
        total_t2i_adapter_embedding,
        total_controlnet_embedding,
        is_sdxl,
        require_inpaint_hijack,
    ):
        x = x.half()
        if y is not None:
            y = y.half()
        context = context.half()
        hs = []
        with flow.no_grad():
            t_emb = cond_cast_unet(
                timestep_embedding(timesteps, self.model_channels, repeat_only=False)
            )
            emb = self.time_embed(t_emb.half())

            if is_sdxl:
                assert y.shape[0] == x.shape[0]
                emb = emb + self.label_emb(y)

            h = x
            for i, module in enumerate(self.input_blocks):
                self.current_h_shape = (h.shape[0], h.shape[1], h.shape[2], h.shape[3])
                h = module(h, emb, context)

                t2i_injection = [3, 5, 8] if is_sdxl else [2, 5, 8, 11]

                if i in t2i_injection:
                    h = aligned_adding(
                        h, total_t2i_adapter_embedding.pop(0), require_inpaint_hijack
                    )

                hs.append(h)

            self.current_h_shape = (h.shape[0], h.shape[1], h.shape[2], h.shape[3])
            h = self.middle_block(h, emb, context)

        # U-Net Middle Block
        h = aligned_adding(h, total_controlnet_embedding.pop(), require_inpaint_hijack)

        if len(total_t2i_adapter_embedding) > 0 and is_sdxl:
            h = aligned_adding(
                h, total_t2i_adapter_embedding.pop(0), require_inpaint_hijack
            )

        # U-Net Decoder
        for i, module in enumerate(self.output_blocks):
            self.current_h_shape = (h.shape[0], h.shape[1], h.shape[2], h.shape[3])
            h = cat(
                [
                    h,
                    aligned_adding(
                        hs.pop(),
                        total_controlnet_embedding.pop(),
                        require_inpaint_hijack,
                    ),
                ],
                dim=1,
            )
            h = h.half()
            h = module(h, emb, context)

        # U-Net Output
        h = h.type(x.dtype)
        h = self.out(h)

        return h



"""
Install:
    pip install pytest
Uasge:
    python -m pytest diffusers/tests/torch_to_oflow/test_temp_fix_compile_impl.py
"""
import numpy as np
import pytest
from onediff.infer_compiler.backends.oneflow.import_tools.patch_for_compiler import (
    FakeCuda,
)


@pytest.mark.parametrize("batch_size", [8])
@pytest.mark.parametrize("n_heads", [4])
@pytest.mark.parametrize("seq_len_q", [4, 8, 64, 128, 256, 512])
@pytest.mark.parametrize("seq_len_k", [4, 8, 64, 128, 256, 512])
@pytest.mark.parametrize("head_dim", [8, 16, 32, 64])
@pytest.mark.parametrize("is_causal", [True, False])
@pytest.mark.parametrize("dropout_p", [0.0])
# @pytest.mark.parametrize("dropout_p", [0.0, 0.22, 0.48])
@pytest.mark.parametrize("seed", [1])
def test_flash_attention(
    batch_size, n_heads, seq_len_q, seq_len_k, head_dim, is_causal, dropout_p, seed
):
    np.random.seed(seed=seed)
    query = np.random.rand(batch_size, n_heads, seq_len_q, head_dim).astype(np.float16)
    key = np.random.rand(batch_size, n_heads, seq_len_k, head_dim).astype(np.float16)
    value = np.random.rand(batch_size, n_heads, seq_len_k, head_dim).astype(np.float16)

    def torch_flash_attention() -> np.ndarray:
        import torch

        q = torch.from_numpy(query).to("cuda")
        k = torch.from_numpy(key).to("cuda")
        v = torch.from_numpy(value).to("cuda")
        result = torch.nn.functional.scaled_dot_product_attention(
            query=q,
            key=k,
            value=v,
            attn_mask=None,
            dropout_p=dropout_p,
            is_causal=is_causal,
        )
        return result.cpu().detach().numpy()

    def oneflow_flash_attention() -> np.ndarray:
        import oneflow as flow  # usort: skip

        q = flow.tensor(query, dtype=flow.float16).to("cuda")
        k = flow.tensor(key, dtype=flow.float16).to("cuda")
        v = flow.tensor(value, dtype=flow.float16).to("cuda")

        return FakeCuda.scaled_dot_product_attention(
            query=q,
            key=k,
            value=v,
            attn_mask=None,
            dropout_p=dropout_p,
            is_causal=is_causal,
        ).numpy()

    torch_result = torch_flash_attention()

    result = oneflow_flash_attention()  # BHMK

    assert np.allclose(torch_result.flatten(), result.flatten(), atol=1e-2, rtol=1e-2)



import os
import time
from contextlib import contextmanager
from typing import Dict

# ComfyUI
import folder_paths
import torch
import torch.nn as nn
from nodes import KSampler, VAEDecode
from onediff.infer_compiler import oneflow_compile

# onediff
from onediff.torch_utils.module_operations import get_sub_module, modify_sub_module
from onediff_quant import Quantizer

# onediff_quant
from onediff_quant.utils import (
    find_quantizable_modules,
    get_quantize_module,
    metric_quantize_costs,
    symm_quantize,
)

# onediff_comfy_nodes
from .model_patcher import OneFlowDeepCacheSpeedUpModelPatcher


def quantize_sub_module(model, sub_name, sub_module, bits):
    if sub_module is None:
        sub_module = get_sub_module(model, sub_name)

    quantizer = Quantizer()
    quantizer.configure(bits=bits, perchannel=True)
    quantizer.find_params(sub_module.weight.float(), weight=True)
    shape = [-1] + [1] * (len(sub_module.weight.shape) - 1)
    scale = quantizer.scale.reshape(*shape)

    org_weight_data = sub_module.weight.data
    org_requires_grad = sub_module.weight.requires_grad

    # save_as_float = False
    sub_module.weight.requires_grad = False
    sub_module.weight.data = symm_quantize(
        sub_module.weight.data, scale.to(sub_module.weight.data.device), quantizer.maxq
    )
    input_scale_and_zero_point = [None, None]
    quant_module = get_quantize_module(
        sub_module,
        sub_name,
        input_scale_and_zero_point + [scale.reshape(-1).tolist()],
        False,  # fake_quant
        False,
        bits,
    )

    modify_sub_module(model, sub_name, quant_module)

    def restore():
        sub_module.weight.data = org_weight_data
        sub_module.weight.requires_grad = org_requires_grad
        modify_sub_module(model, sub_name, sub_module)

    return restore


class quantized_model_patcher:
    def __init__(self, model_patcher, layers, bits, verbose=False):
        self.model_patcher = model_patcher
        self.diffusion_model = model_patcher.model.diffusion_model
        self.layers = layers
        self.bits = bits
        self.handles = []
        self.verbose = verbose
        self.conv_count = 0
        self.linear_count = 0

    def __enter__(self):
        self.start_time = time.time()
        diffusion_model = self.model_patcher.model.diffusion_model
        for sub_name in self.layers:
            sub_module = get_sub_module(diffusion_model, sub_name)
            if isinstance(sub_module, nn.Conv2d):
                self.conv_count += 1
            elif isinstance(sub_module, nn.Linear):
                self.linear_count += 1

            restore = quantize_sub_module(
                diffusion_model, sub_name, sub_module, bits=self.bits
            )
            self.handles.append(restore)

        return self.model_patcher

    def __exit__(self, exc_type, exc_val, exc_tb):
        for handle in self.handles:
            handle()

        if self.verbose:
            end = time.time()
            print(f"\n=> Total time: {end - self.start_time:.4f} seconds")
            print(f"Quantized {self.conv_count} convs, {self.linear_count} linears.\n")
        return False


class KSampleQuantumBase(KSampler):
    @classmethod
    def INPUT_TYPES(s):
        ret = KSampler.INPUT_TYPES()

        ret["required"].update(
            {"onediff_quant": ("STRING", {"default": "Quantization Settings"})}
        )
        ret["required"].update({"bits": ("INT", {"default": 8, "min": 8, "max": 8})})
        ret["required"].update({"quantize_conv": (["enable", "disable"],)})
        ret["required"].update({"quantize_linear": (["enable", "disable"],)})
        return ret

    CATEGORY = "OneDiff/Quant_Tools"

    def generate_latent_sample(self, model_patcher, *args, **kwargs):
        # self.sample from KSampler
        latent_samples = self.sample(model_patcher, *args, **kwargs)
        assert (
            len(latent_samples) == 1
        ), "Only batch size 1 is supported for quantization"
        out = latent_samples[0]
        return out

    def generate_pipeline(self, model_patcher, *args, **kwargs):
        generate_img = self.generate_latent_sample

        class ImageGenerationPipeline:
            def __init__(self):
                self.model = model_patcher
                self.args = args
                self.kwargs = kwargs

            def __call__(self, *args, **kwargs):
                return generate_img(self.model, *self.args, **self.kwargs)

        return ImageGenerationPipeline()

    def generate_calibrate_info(
        self,
        model_patcher,
        only_compute_density: bool = False,
        bits=8,
        quantized_model_generator: callable = lambda x: x.model.diffusion_model,
        model_cls=[nn.Linear, nn.Conv2d],
        *args,
        **kwargs,
    ) -> Dict:
        """return calibrate_info"""

        calibrate_info = {}

        diffusion_model = quantized_model_generator(model_patcher)

        quantizable_modules = find_quantizable_modules(
            diffusion_model, module_cls=model_cls
        )

        org_latent_sample = self.generate_latent_sample(model_patcher, *args, **kwargs)[
            "samples"
        ]
        pipe = self.generate_pipeline(model_patcher, *args, **kwargs)
        quantize_costs = metric_quantize_costs(
            pipe, pipe_kwargs={}, quantizable_modules=quantizable_modules
        )

        if only_compute_density:
            for sub_name, _ in quantizable_modules.items():
                compute_density = quantize_costs.get_compute_density(sub_name)
                calibrate_info[sub_name] = {
                    "compute_density": compute_density,
                }

            return calibrate_info

        length = len(quantizable_modules)
        start_time = time.time()
        for index, (sub_name, _) in enumerate(quantizable_modules.items()):
            print(f"Quantizing {index+1}/{length} {sub_name}...")

            with quantized_model_patcher(
                model_patcher=model_patcher, layers=[sub_name], bits=bits
            ) as qmpatcher:
                cur_latent_sample = self.generate_latent_sample(
                    qmpatcher, *args, **kwargs
                )["samples"]

                mse = torch.mean((org_latent_sample - cur_latent_sample) ** 2)

                compute_density = quantize_costs.get_compute_density(sub_name)
                print(f"mse: {mse:.4f}, compute_density: {compute_density:.4f}")

                calibrate_info[sub_name] = {
                    "mse": mse,
                    "compute_density": compute_density,
                }

            estimated_remaining_time = (
                (time.time() - start_time) / (index + 1) * (length - index - 1) / 60
            )
            print(f"Estimated remaining time: {estimated_remaining_time:.4f} minutes")

        return calibrate_info


class FineTuneCalibrateInfoMixin:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "conv_mse_threshold": (
                    "FLOAT",
                    {
                        "default": 0.1,
                        "min": 0,
                        "max": 1,
                        "step": 0.01,
                        "display": "number",
                    },
                ),
                "linear_mse_threshold": (
                    "FLOAT",
                    {
                        "default": 0.1,
                        "min": 0,
                        "max": 1,
                        "step": 0.01,
                        "display": "number",
                    },
                ),
                "compute_density_threshold": (
                    "INT",
                    {
                        "default": 0,
                        "min": 0,
                        "max": 100000,
                        "step": 1,
                        "display": "number",
                    },
                ),
            },
        }

    def fine_tune_calibrate_info(
        self,
        model,
        calibrate_info,
        conv_mse_threshold,
        linear_mse_threshold,
        compute_density_threshold,
    ):
        new_calibrate_info = {}
        for sub_name, sub_info in calibrate_info.items():
            if sub_info["compute_density"] < compute_density_threshold:
                continue

            sub_model = get_sub_module(model.model.diffusion_model, sub_name)

            if isinstance(sub_model, nn.Conv2d) and "mse" in sub_info:
                if sub_info["mse"] > conv_mse_threshold:
                    continue

            if isinstance(sub_model, nn.Linear) and "mse" in sub_info:
                if sub_info["mse"] > linear_mse_threshold:
                    continue

            new_calibrate_info[sub_name] = sub_info

        return new_calibrate_info


class SaveQuantizedCalibrateInfoMixin:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "save_filename_prefix": ("STRING", {"default": "unet"}),
                "overwrite": (["disable", "enable"],),
            },
        }

    def save_quantized_calibrate_info(
        self, quantize_config_file, overwrite, calibrate_info
    ):
        if not overwrite:
            if quantize_config_file.exists():
                print(f"Quantize config file {quantize_config_file} exists, skip.")
                return

        torch.save(calibrate_info, quantize_config_file)
        print(f"Saved quantize config file to {quantize_config_file}")

        return quantize_config_file


@contextmanager
def compile_model_patcher_context(model_patcher):
    if isinstance(model_patcher, OneFlowDeepCacheSpeedUpModelPatcher):
        original_models = {
            "deep_cache_unet": model_patcher.deep_cache_unet,
            "fast_deep_cache_unet": model_patcher.fast_deep_cache_unet,
        }
        model_patcher.deep_cache_unet = oneflow_compile(model_patcher.deep_cache_unet)
        model_patcher.fast_deep_cache_unet = oneflow_compile(
            model_patcher.fast_deep_cache_unet
        )
    else:
        original_diffusion_model = model_patcher.model.diffusion_model
        model_patcher.model.diffusion_model = oneflow_compile(
            model_patcher.model.diffusion_model
        )

    yield model_patcher

    if isinstance(model_patcher, OneFlowDeepCacheSpeedUpModelPatcher):
        model_patcher.deep_cache_unet = original_models["deep_cache_unet"]
        model_patcher.fast_deep_cache_unet = original_models["fast_deep_cache_unet"]
    else:
        model_patcher.model.diffusion_model = original_diffusion_model



import os

import comfy
import torch
import torch.nn as nn

if hasattr(comfy.ops, "disable_weight_init"):
    comfy_ops_Linear = comfy.ops.disable_weight_init.Linear
else:
    print(
        "Warning: ComfyUI version is too old, please upgrade it. github: git@github.com:comfyanonymous/ComfyUI.git "
    )
    comfy_ops_Linear = comfy.ops.Linear

__all__ = ["replace_module_with_quantizable_module"]


def get_sub_module(module, sub_module_name) -> nn.Module:
    """Get a submodule of a module using dot-separated names.

    Args:
        module (nn.Module): The base module.
        sub_module_name (str): Dot-separated name of the submodule.

    Returns:
        nn.Module: The requested submodule.
    """

    parts = sub_module_name.split(".")
    current_module = module

    for part in parts:
        try:
            if part.isdigit():
                current_module = current_module[int(part)]
            else:
                current_module = getattr(current_module, part)
        except (IndexError, AttributeError):
            raise ModuleNotFoundError(f"Submodule {part} not found.")

    return current_module


def modify_sub_module(module, sub_module_name, new_value):
    """Modify a submodule of a module using dot-separated names.

    Args:
        module (nn.Module): The base module.
        sub_module_name (str): Dot-separated name of the submodule.
        new_value: The new value to assign to the submodule.

    """
    parts = sub_module_name.split(".")
    current_module = module

    for i, part in enumerate(parts):
        try:
            if part.isdigit():
                if i == len(parts) - 1:
                    current_module[int(part)] = new_value
                else:
                    current_module = current_module[int(part)]
            else:
                if i == len(parts) - 1:
                    setattr(current_module, part, new_value)
                else:
                    current_module = getattr(current_module, part)
        except (IndexError, AttributeError):
            raise ModuleNotFoundError(f"Submodule {part} not found.")


def _load_calibrate_info(calibrate_info_path):
    calibrate_info = {}
    with open(calibrate_info_path, "r") as f:
        for line in f.readlines():
            line = line.strip()
            items = line.split(" ")
            calibrate_info[items[0]] = [
                float(items[1]),
                int(items[2]),
                [float(x) for x in items[3].split(",")],
            ]
    return calibrate_info


def search_modules(root, match_fn: callable, name=""):
    """
    example:
    >>> search_modules(model, lambda m: isinstance(m, (nn.Conv2d, nn.Linear))
    """
    if match_fn(root):
        return {name: root}

    result = {}
    for child_name, child in root.named_children():
        result.update(
            search_modules(
                child, match_fn, f"{name}.{child_name}" if name != "" else child_name
            )
        )
    return result


def _can_use_flash_attn(attn):
    dim_head = attn.to_q.out_features // attn.heads
    if dim_head != 40 and dim_head != 64:
        return False
    if attn.to_k is None or attn.to_v is None:
        return False
    if (
        attn.to_q.bias is not None
        or attn.to_k.bias is not None
        or attn.to_v.bias is not None
    ):
        return False
    if (
        attn.to_q.in_features != attn.to_k.in_features
        or attn.to_q.in_features != attn.to_v.in_features
    ):
        return False
    if not (
        attn.to_q.weight.dtype == attn.to_k.weight.dtype
        and attn.to_q.weight.dtype == attn.to_v.weight.dtype
    ):
        return False
    return True


def _rewrite_attention(attn):
    from onediff_quant.models import DynamicQuantLinearModule, StaticQuantLinearModule

    dim_head = attn.to_q.out_features // attn.heads
    has_bias = attn.to_q.bias is not None
    attn.to_qkv = nn.Linear(
        attn.to_q.in_features, attn.to_q.out_features * 3, bias=has_bias
    )
    attn.to_qkv.requires_grad_(False)
    qkv_weight = torch.cat(
        [
            attn.to_q.weight.permute(1, 0).reshape(-1, attn.heads, dim_head),
            attn.to_k.weight.permute(1, 0).reshape(-1, attn.heads, dim_head),
            attn.to_v.weight.permute(1, 0).reshape(-1, attn.heads, dim_head),
        ],
        dim=2,
    )
    qkv_weight = (
        qkv_weight.reshape(-1, attn.to_q.out_features * 3).permute(1, 0).contiguous()
    )
    attn.to_qkv.weight.data = qkv_weight

    if has_bias:
        qkv_bias = (
            torch.cat(
                [
                    attn.to_q.bias.reshape(attn.heads, dim_head),
                    attn.to_k.bias.reshape(attn.heads, dim_head),
                    attn.to_v.bias.reshape(attn.heads, dim_head),
                ],
                dim=1,
            )
            .reshape(attn.to_q.out_features * 3)
            .contiguous()
        )
        attn.to_qkv.bias.data = qkv_bias

    if isinstance(attn.to_q, StaticQuantLinearModule) or isinstance(
        attn.to_q, DynamicQuantLinearModule
    ):
        cls = type(attn.to_q)
        weight_scale = (
            torch.cat(
                [
                    torch.Tensor(attn.to_q.calibrate[2]).reshape(attn.heads, dim_head),
                    torch.Tensor(attn.to_k.calibrate[2]).reshape(attn.heads, dim_head),
                    torch.Tensor(attn.to_v.calibrate[2]).reshape(attn.heads, dim_head),
                ],
                dim=1,
            )
            .reshape(attn.to_q.out_features * 3)
            .contiguous()
        )
        calibrate = [attn.to_q.calibrate[0], attn.to_q.calibrate[1], weight_scale]

        old_env = os.getenv("ONEFLOW_FUSE_QUANT_TO_MATMUL")
        os.environ["ONEFLOW_FUSE_QUANT_TO_MATMUL"] = "0"
        attn.to_qkv = cls(attn.to_qkv, attn.to_q.nbits, calibrate, attn.to_q.name)
        attn.scale = dim_head**-0.5

        os.environ["ONEFLOW_FUSE_QUANT_TO_MATMUL"] = old_env


def replace_module_with_quantizable_module(
    diffusion_model, calibrate_info_path, use_rewrite_attn=True
):
    from onediff_quant.utils import get_quantize_module

    calibrate_info = _load_calibrate_info(calibrate_info_path)
    for sub_module_name, sub_calibrate_info in calibrate_info.items():
        sub_mod = get_sub_module(diffusion_model, sub_module_name)

        if isinstance(sub_mod, comfy_ops_Linear):
            # fix onediff_quant use isinstance(sub_mod, torch.nn.Linear)
            sub_mod.__class__ = torch.nn.Linear

        sub_mod.weight.requires_grad = False
        sub_mod.weight.data = sub_mod.weight.to(torch.int8)
        sub_mod.cuda()  # TODO: remove this line , because we onediff_quant pkg weight_scale
        sub_mod = get_quantize_module(
            sub_mod,
            sub_module_name,
            sub_calibrate_info,
            fake_quant=False,
            static=False,
            nbits=8,
        )
        modify_sub_module(diffusion_model, sub_module_name, sub_mod)
    if use_rewrite_attn:
        print(f"{use_rewrite_attn=}, rewrite CrossAttention")
        try:
            # rewrite CrossAttention to use qkv
            from comfy.ldm.modules.attention import CrossAttention

            match_func = lambda m: isinstance(
                m, CrossAttention
            ) and _can_use_flash_attn(m)
            can_rewrite_modules = search_modules(diffusion_model, match_func)
            print(f"rewrite {len(can_rewrite_modules)=} CrossAttention")
            for k, v in can_rewrite_modules.items():
                if f"{k}.to_q" in calibrate_info:
                    _rewrite_attention(v)  # diffusion_model is modified in-place
                else:
                    print(f"skip {k+'.to_q'} not in calibrate_info")

        except Exception as e:
            raise RuntimeError(f"rewrite CrossAttention failed: {e}")


def find_quantizable_modules(
    module, name="", *, quantize_conv=True, quantize_linear=True
):
    if isinstance(module, nn.Conv2d) and quantize_conv:
        return {name: module}

    if isinstance(module, nn.Linear) and quantize_linear:
        return {name: module}

    res = {}
    for child_name, child in module.named_children():
        res.update(
            find_quantizable_modules(
                child,
                name + "." + child_name if name != "" else child_name,
                quantize_conv=quantize_conv,
                quantize_linear=quantize_linear,
            )
        )
    return res


def quantize_and_save_model(
    diffusion_model,
    output_dir,
    *,
    quantize_conv=True,
    quantize_linear=True,
    verbose=False,
    bits=8,
):
    import time

    from onediff_quant import Quantizer
    from onediff_quant.utils import symm_quantize_sub_module
    from safetensors.torch import save_model

    print(
        f"quantize and save_model, conv={quantize_conv}, linear={quantize_linear}, verbose={verbose}, output={output_dir}"
    )
    start_time = time.time()

    print("Find the quantizable modules...")
    quantizable_modules = find_quantizable_modules(
        diffusion_model, quantize_conv=quantize_conv, quantize_linear=quantize_linear
    )
    print(f"quantizable_modules size: {len(quantizable_modules)}")

    if verbose:
        for name, module in quantizable_modules.items():
            print(name, ": ", module)

    calibrate_info = {}

    enum_quantizable_modules = enumerate(quantizable_modules.items())
    _quant_total = len(quantizable_modules.items())
    for i, (name, module) in enum_quantizable_modules:
        if verbose:
            print(f"Calculate quantization infos of {name} ...")
        quantizer = Quantizer()
        quantizer.configure(bits=bits, perchannel=True)
        quantizer.find_params(module.weight.float(), weight=True)
        shape = [-1] + [1] * (len(module.weight.shape) - 1)
        scale = quantizer.scale.reshape(*shape)
        symm_quantize_sub_module(diffusion_model, name, scale, quantizer.maxq)
        calibrate_info[name] = [scale]

    print(f"""save quantized model to {output_dir}""")
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    else:
        raise RuntimeError(
            f"{os.path.basename(output_dir)} has existed, rename the out_dir and try again."
        )
    save_model(diffusion_model, os.path.join(output_dir, "unet_int8.safetensors"))

    print(f'save calibrate_info to {os.path.join(output_dir, "calibrate_info.txt")}')

    with open(os.path.join(output_dir, "calibrate_info.txt"), "w") as f:
        for name, info in calibrate_info.items():
            input_scale = 0
            input_zero_point = 0
            weight_scale = [str(x) for x in info[0].reshape(-1).tolist()]
            f.write(
                f"{name} {input_scale} {input_zero_point} {','.join(weight_scale)}\n"
            )

    print(f"Quantize module time: {time.time() - start_time}s")



import os
from pathlib import Path

import numpy as np
import pytest
from PIL import Image
from utils import (
    cal_ssim,
    check_and_generate_images,
    dump_image,
    get_all_args,
    get_base_args,
    get_data_summary,
    get_image_array_from_response,
    get_target_image_filename,
    get_threshold,
    IMG2IMG_API_ENDPOINT,
    is_txt2img,
    OPTIONS_API_ENDPOINT,
    post_request_and_check,
    SAVED_GRAPH_NAME,
    TXT2IMG_API_ENDPOINT,
    WEBUI_SERVER_URL,
)


@pytest.fixture(scope="session", autouse=True)
def change_model():
    option_payload = {
        "sd_model_checkpoint": "checkpoints/AWPainting_v1.2.safetensors",
    }
    post_request_and_check(f"{WEBUI_SERVER_URL}/{OPTIONS_API_ENDPOINT}", option_payload)


@pytest.fixture(scope="session", autouse=True)
def prepare_target_images():
    print("checking if target images exist...")
    check_and_generate_images()


@pytest.fixture()
def base_url():
    return WEBUI_SERVER_URL


@pytest.fixture()
def url_txt2img(base_url):
    return f"{base_url}/{TXT2IMG_API_ENDPOINT}"


@pytest.fixture()
def url_img2img(base_url):
    return f"{base_url}/{IMG2IMG_API_ENDPOINT}"


@pytest.fixture()
def url_set_config(base_url):
    return f"{base_url}/{OPTIONS_API_ENDPOINT}"


@pytest.mark.parametrize("data", get_all_args())
def test_image_ssim(base_url, data):
    print(f"testing: {get_data_summary(data)}")
    endpoint = TXT2IMG_API_ENDPOINT if is_txt2img(data) else IMG2IMG_API_ENDPOINT
    url = f"{base_url}/{endpoint}"
    generated_image = get_image_array_from_response(post_request_and_check(url, data))
    target_image_path = get_target_image_filename(data)
    target_image = np.array(Image.open(target_image_path))
    ssim_value = cal_ssim(generated_image, target_image)
    if ssim_value < get_threshold(data):
        dump_image(target_image, generated_image, Path(target_image_path).name)
    assert ssim_value > get_threshold(data)


def test_onediff_save_graph(url_txt2img):
    script_args = {
        "script_args": [
            False,  # quantization
            None,  # graph_checkpoint
            SAVED_GRAPH_NAME,  # saved_graph_name
        ]
    }
    data = {**get_base_args(), **script_args}
    post_request_and_check(url_txt2img, data)


def test_onediff_load_graph(url_txt2img):
    script_args = {
        "script_args": [
            False,  # quantization
            SAVED_GRAPH_NAME,  # graph_checkpoint
            "",  # saved_graph_name
        ]
    }
    data = {**get_base_args(), **script_args}
    post_request_and_check(url_txt2img, data)


@pytest.mark.skip
def test_onediff_refiner(url_txt2img):
    extra_args = {
        "sd_model_checkpoint": "sd_xl_base_1.0.safetensors",
        "refiner_checkpoint": "sd_xl_refiner_1.0.safetensors [7440042bbd]",
        "refiner_switch_at": 0.8,
    }
    data = {**get_base_args(), **extra_args}
    # loop 3 times for checking model switching between base and refiner
    for _ in range(3):
        post_request_and_check(url_txt2img, data)



import math
from inspect import isfunction

import oneflow as flow  # usort: skip
from oneflow import nn


# https://github.com/Stability-AI/generative-models/blob/059d8e9cd9c55aea1ef2ece39abf605efb8b7cc9/sgm/modules/diffusionmodules/util.py#L274
# https://github.com/Stability-AI/stablediffusion/blob/b4bdae9916f628461e1e4edbc62aafedebb9f7ed/ldm/modules/diffusionmodules/util.py#L224
class GroupNorm32Oflow(nn.GroupNorm):
    def forward(self, x):
        return super().forward(x)


# https://github.com/Stability-AI/generative-models/blob/059d8e9cd9c55aea1ef2ece39abf605efb8b7cc9/sgm/modules/diffusionmodules/util.py#L207
def timestep_embedding(timesteps, dim, max_period=10000, repeat_only=False):
    if not repeat_only:
        half = dim // 2
        freqs = flow.exp(
            -math.log(max_period)
            * flow.arange(start=0, end=half, dtype=flow.float32)
            / half
        ).to(device=timesteps.device)
        args = timesteps[:, None].float() * freqs[None]
        embedding = flow.cat([flow.cos(args), flow.sin(args)], dim=-1)
        if dim % 2:
            embedding = flow.cat([embedding, flow.zeros_like(embedding[:, :1])], dim=-1)
    else:
        raise NotImplementedError(
            "repeat_only=True is not implemented in timestep_embedding"
        )
    return embedding


def exists(val):
    return val is not None


def default(val, d):
    if exists(val):
        return val
    return d() if isfunction(d) else d


# https://github.com/AUTOMATIC1111/stable-diffusion-webui/blob/master/modules/sd_hijack_optimizations.py#L142
# https://github.com/AUTOMATIC1111/stable-diffusion-webui/blob/master/modules/sd_hijack_optimizations.py#L221
class CrossAttentionOflow(nn.Module):
    def __init__(
        self,
        query_dim,
        context_dim=None,
        heads=8,
        dim_head=64,
        dropout=0.0,
        backend=None,
    ):
        super().__init__()
        inner_dim = dim_head * heads
        context_dim = default(context_dim, query_dim)

        self.scale = dim_head**-0.5
        self.heads = heads

        self.to_q = nn.Linear(query_dim, inner_dim, bias=False)
        self.to_k = nn.Linear(context_dim, inner_dim, bias=False)
        self.to_v = nn.Linear(context_dim, inner_dim, bias=False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, query_dim), nn.Dropout(dropout)
        )
        self.backend = backend

    def forward(
        self,
        x,
        context=None,
        mask=None,
        additional_tokens=None,
        n_times_crossframe_attn_in_self=0,
    ):
        h = self.heads

        q_in = self.to_q(x)
        context = default(context, x)

        # context_k, context_v = hypernetwork.apply_hypernetworks(shared.loaded_hypernetworks, context)
        context_k, context_v = context, context
        k_in = self.to_k(context_k)
        v_in = self.to_v(context_v)
        out = flow._C.fused_multi_head_attention_inference_v2(
            query=q_in,
            query_layout="BM(HK)",
            query_head_size=self.to_q.out_features // self.heads,
            key=k_in,
            key_layout="BM(HK)",
            value=v_in,
            value_layout="BM(HK)",
            output_layout="BM(HK)",
            causal=False,
        )
        return self.to_out(out)



"""hijack ComfyUI/comfy/samplers.py
commit: 4bd7d55b9028d79829a645edfe8259f7b7a049c0
Date:   Thu Apr 11 22:43:05 2024 -0400
"""

from typing import Dict

import torch
from comfy.samplers import calc_cond_batch, can_concat_cond, cond_cat, get_area_and_mult

from ..sd_hijack_utils import Hijacker
from .booster_utils import is_using_nexfort_backend
from .patch_management import create_patch_executor, PatchType


def calc_cond_batch_of(orig_func, model, conds, x_in, timestep, model_options):
    out_conds = []
    out_counts = []
    to_run = []

    for i in range(len(conds)):
        out_conds.append(torch.zeros_like(x_in))
        out_counts.append(torch.ones_like(x_in) * 1e-37)

        cond = conds[i]
        if cond is not None:
            for x in cond:
                p = get_area_and_mult(x, x_in, timestep)
                if p is None:
                    continue

                to_run += [(p, i)]

    while len(to_run) > 0:
        first = to_run[0]
        first_shape = first[0][0].shape
        to_batch_temp = []
        for x in range(len(to_run)):
            if can_concat_cond(to_run[x][0], first[0]):
                to_batch_temp += [x]

        to_batch_temp.reverse()
        # to_batch = to_batch_temp[:1]
        to_batch = to_batch_temp
        # free_memory = model_management.get_free_memory(x_in.device)
        # for i in range(1, len(to_batch_temp) + 1):
        #     batch_amount = to_batch_temp[:len(to_batch_temp)//i]
        #     input_shape = [len(batch_amount) * first_shape[0]] + list(first_shape)[1:]
        #     if model.memory_required(input_shape) < free_memory:
        #         to_batch = batch_amount
        #         break

        input_x = []
        mult = []
        c = []
        cond_or_uncond = []
        area = []
        control = None
        patches = None
        for x in to_batch:
            o = to_run.pop(x)
            p = o[0]
            input_x.append(p.input_x)
            mult.append(p.mult)
            c.append(p.conditioning)
            area.append(p.area)
            cond_or_uncond.append(o[1])
            control = p.control
            patches = p.patches

        batch_chunks = len(cond_or_uncond)
        input_x = torch.cat(input_x)
        c = cond_cat(c)
        timestep_ = torch.cat([timestep] * batch_chunks)

        if control is not None:
            c["control"] = control.get_control(
                input_x, timestep_, c, len(cond_or_uncond)
            )

        transformer_options = {}
        if "transformer_options" in model_options:
            transformer_options = model_options["transformer_options"].copy()

        if patches is not None:
            if "patches" in transformer_options:
                cur_patches = transformer_options["patches"].copy()
                for p in patches:
                    if p in cur_patches:
                        cur_patches[p] = cur_patches[p] + patches[p]
                    else:
                        cur_patches[p] = patches[p]
                transformer_options["patches"] = cur_patches
            else:
                transformer_options["patches"] = patches

        transformer_options["cond_or_uncond"] = cond_or_uncond[:]

        diff_model = model.diffusion_model
        transformer_options["sigmas"] = timestep

        if create_patch_executor(PatchType.CachedCrossAttentionPatch).check_patch(
            diff_model
        ):
            patch_executor = create_patch_executor(PatchType.UNetExtraInputOptions)
            extra_options = transformer_options
            sigma = (
                extra_options["sigmas"][0].item()
                if "sigmas" in extra_options
                else 999999999.9
            )
            assert "_sigmas" not in extra_options
            extra_options["_sigmas"] = {}
            attn2_patch_dict = extra_options["patches_replace"]["attn2"]
            for k, attn_m in attn2_patch_dict.items():
                out_lst = []
                for i, callback in enumerate(attn_m.callback):
                    if (
                        sigma <= attn_m.kwargs[i]["sigma_start"]
                        and sigma >= attn_m.kwargs[i]["sigma_end"]
                    ):
                        out_lst.append(1)
                    else:
                        out_lst.append(0)
                # extra inputs
                transformer_options["_sigmas"][attn_m.forward_patch_key] = torch.randn(
                    *out_lst
                )

            # extra inputs
            transformer_options["_attn2"] = patch_executor.get_patch(diff_model)[
                "attn2"
            ]

        c["transformer_options"] = transformer_options
        if "model_function_wrapper" in model_options:
            output = model_options["model_function_wrapper"](
                model.apply_model,
                {
                    "input": input_x,
                    "timestep": timestep_,
                    "c": c,
                    "cond_or_uncond": cond_or_uncond,
                },
            ).chunk(batch_chunks)
        else:
            output = model.apply_model(input_x, timestep_, **c).chunk(batch_chunks)

        for o in range(batch_chunks):
            cond_index = cond_or_uncond[o]
            a = area[o]
            if a is None:
                out_conds[cond_index] += output[o] * mult[o]
                out_counts[cond_index] += mult[o]
            else:
                out_c = out_conds[cond_index]
                out_cts = out_counts[cond_index]
                dims = len(a) // 2
                for i in range(dims):
                    out_c = out_c.narrow(i + 2, a[i + dims], a[i])
                    out_cts = out_cts.narrow(i + 2, a[i + dims], a[i])
                out_c += output[o] * mult[o]
                out_cts += mult[o]

    for i in range(len(out_conds)):
        out_conds[i] /= out_counts[i]

    return out_conds

    #     for o in range(batch_chunks):
    #         cond_index = cond_or_uncond[o]
    #         out_conds[cond_index][
    #             :,
    #             :,
    #             area[o][2] : area[o][0] + area[o][2],
    #             area[o][3] : area[o][1] + area[o][3],
    #         ] += (output[o] * mult[o])
    #         out_counts[cond_index][
    #             :,
    #             :,
    #             area[o][2] : area[o][0] + area[o][2],
    #             area[o][3] : area[o][1] + area[o][3],
    #         ] += mult[o]

    # for i in range(len(out_conds)):
    #     out_conds[i] /= out_counts[i]

    # return out_conds


def cond_func(orig_func, model, *args, **kwargs):
    return is_using_nexfort_backend(model)


samplers_hijack = Hijacker()
samplers_hijack.register(
    orig_func=calc_cond_batch,
    sub_func=calc_cond_batch_of,
    cond_func=cond_func,
)



import json
import os

# Prompt Styler, a custom node for ComfyUI


def read_json_file(file_path):
    """
    Reads a JSON file's content and returns it.
    Ensures content matches the expected format.
    """
    if not os.access(file_path, os.R_OK):
        print(f"Warning: No read permissions for file {file_path}")
        return None

    try:
        with open(file_path, "r", encoding="utf-8") as file:
            content = json.load(file)
            # Check if the content matches the expected format.
            if not all(
                [
                    "name" in item and "prompt" in item and "negative_prompt" in item
                    for item in content
                ]
            ):
                print(f"Warning: Invalid content in file {file_path}")
                return None
            return content
    except Exception as e:
        print(f"An error occurred while reading {file_path}: {str(e)}")
        return None


def read_sdxl_styles(json_data):
    """
    Returns style names from the provided JSON data.
    """
    if not isinstance(json_data, list):
        print("Error: input data must be a list")
        return []

    return [
        item["name"] for item in json_data if isinstance(item, dict) and "name" in item
    ]


def get_all_json_files(directory):
    return [
        os.path.join(directory, file)
        for file in os.listdir(directory)
        if file.endswith(".json") and os.path.isfile(os.path.join(directory, file))
    ]


def load_styles_from_directory(directory):
    """
    Loads styles from all JSON files in the directory.
    Renames duplicate style names by appending a suffix.
    """
    json_files = get_all_json_files(directory)
    combined_data = []
    seen = set()

    for json_file in json_files:
        json_data = read_json_file(json_file)
        if json_data:
            for item in json_data:
                original_style = item["name"]
                style = original_style
                suffix = 1
                while style in seen:
                    style = f"{original_style}_{suffix}"
                    suffix += 1
                item["name"] = style
                seen.add(style)
                combined_data.append(item)

    unique_style_names = [
        item["name"]
        for item in combined_data
        if isinstance(item, dict) and "name" in item
    ]

    return combined_data, unique_style_names


def find_template_by_name(json_data, template_name):
    """
    Returns a template from the JSON data by name or None if not found.
    """
    for template in json_data:
        if template["name"] == template_name:
            return template
    return None


class CLIPTextEncodePromptStyle:
    @classmethod
    def INPUT_TYPES(s):
        style_dir = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "style_template"
        )
        s.json_data, styles = load_styles_from_directory(style_dir)
        return {
            "required": {
                "text_positive": (
                    "STRING",
                    {"multiline": True, "dynamicPrompts": True},
                ),
                "text_negative": (
                    "STRING",
                    {"multiline": True, "dynamicPrompts": True},
                ),
                "style": ((styles),),
                "clip": ("CLIP",),
            },
            "optional": {
                "log_prompt": (
                    "BOOLEAN",
                    {"default": True, "label_on": "yes", "label_off": "no"},
                ),
                "style_positive": (
                    "BOOLEAN",
                    {"default": True, "label_on": "yes", "label_off": "no"},
                ),
                "style_negative": (
                    "BOOLEAN",
                    {"default": True, "label_on": "yes", "label_off": "no"},
                ),
            },
        }

    RETURN_TYPES = (
        "CONDITIONING",
        "CONDITIONING",
    )
    RETURN_NAMES = (
        "positive",
        "negative",
    )

    FUNCTION = "encode"

    CATEGORY = "OneDiff/Conditioning"

    def encode(
        self,
        clip,
        text_positive,
        text_negative,
        style,
        log_prompt,
        style_positive,
        style_negative,
    ):
        template = find_template_by_name(self.json_data, style)

        if style_positive:
            positive_prompt = template["prompt"].replace("{prompt}", text_positive)
        else:
            positive_prompt = text_positive

        positive_tokens = clip.tokenize(positive_prompt)
        positive_cond, positive_pooled = clip.encode_from_tokens(
            positive_tokens, return_pooled=True
        )

        if style_negative:
            negative_prompt_template = template.get("negative_prompt", "")
            negative_prompt = (
                f"{negative_prompt_template}, {text_negative}"
                if negative_prompt_template and text_negative
                else text_negative or negative_prompt_template
            )
        else:
            negative_prompt = text_negative
        negative_tokens = clip.tokenize(negative_prompt)
        negative_cond, negative_pooled = clip.encode_from_tokens(
            negative_tokens, return_pooled=True
        )

        if log_prompt:
            print(f"{positive_prompt=}\n{negative_prompt=}")
        return (
            [[positive_cond, {"pooled_output": positive_pooled}]],
            [[negative_cond, {"pooled_output": negative_pooled}]],
        )


NODE_CLASS_MAPPINGS = {
    "CLIPTextEncodePromptStyle": CLIPTextEncodePromptStyle,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "CLIPTextEncodePromptStyle": "CLIP Text Encode(PromptStyle)"
}



from enum import Enum

from .patch_executor import (
    CachedCrossAttentionPatch,
    DeepCacheUNetExecutorPatch,
    UiNodeWithIndexPatch,
    UNetExtraInputOptions,
)


class PatchType(Enum):
    CachedCrossAttentionPatch = CachedCrossAttentionPatch
    UNetExtraInputOptions = UNetExtraInputOptions
    DCUNetExecutorPatch = DeepCacheUNetExecutorPatch
    UiNodeWithIndexPatch = UiNodeWithIndexPatch


def create_patch_executor(selected_patch_type):
    for patch_type in PatchType:
        if selected_patch_type == patch_type:
            return patch_type.value()
    raise NotImplementedError(selected_patch_type)


if __name__ == "__main__":
    patch_executor = create_patch_executor(PatchType.CachedCrossAttentionPatch)
    print(patch_executor)



"""hijack ComfyUI/custom_nodes/ComfyUI_IPAdapter_plus/IPAdapterPlus.py"""
from ..booster_utils import is_using_nexfort_backend
from ._config import ipadapter_plus, ipadapter_plus_hijacker
from .set_model_patch_replace import set_model_patch_replace

set_model_patch_replace_fn = ipadapter_plus.IPAdapterPlus.set_model_patch_replace


def cond_func(org_fn, model, *args, **kwargs):
    return is_using_nexfort_backend(model)


ipadapter_plus_hijacker.register(
    set_model_patch_replace_fn, set_model_patch_replace, cond_func
)



import torch
from comfy.ldm.modules.diffusionmodules.openaimodel import (
    apply_control,
    forward_timestep_embed,
)
from comfy.ldm.modules.diffusionmodules.util import timestep_embedding
from torch.nn import Module


class DeepCacheUNet(Module):
    def __init__(self, unet_module, cache_layer_id, cache_block_id):
        super().__init__()
        self.unet_module = unet_module
        cache_depth = 0
        for i in range(cache_layer_id):
            cache_depth = cache_depth + self.unet_module.num_res_blocks[i]
        cache_depth = cache_depth + cache_block_id + 1
        self.cache_depth = cache_depth

    def forward(
        self,
        x,
        timesteps=None,
        context=None,
        y=None,
        control=None,
        transformer_options={},
        **kwargs
    ):
        """
        Apply the model to an input batch.
        :param x: an [N x C x ...] Tensor of inputs.
        :param timesteps: a 1-D batch of timesteps.
        :param context: conditioning plugged in via crossattn
        :param y: an [N] Tensor of labels, if class-conditional.
        :return: an [N x C x ...] Tensor of outputs.
        """
        transformer_options["original_shape"] = list(x.shape)
        transformer_options["transformer_index"] = 0
        transformer_patches = transformer_options.get("patches", {})

        c_dict = kwargs.get("c", {})
        num_video_frames = c_dict.get(
            "num_video_frames", self.unet_module.default_num_video_frames
        )

        default_image_only_indicator = getattr(
            self.unet_module, "default_image_only_indicator", None
        )
        image_only_indicator = c_dict.get(
            "image_only_indicator", default_image_only_indicator
        )

        time_context = kwargs.get("time_context", None)

        assert (y is not None) == (
            self.unet_module.num_classes is not None
        ), "must specify y if and only if the model is class-conditional"
        hs = []
        t_emb = timestep_embedding(
            timesteps, self.unet_module.model_channels, repeat_only=False
        ).to(x.dtype)
        emb = self.unet_module.time_embed(t_emb)

        if self.unet_module.num_classes is not None:
            assert y.shape[0] == x.shape[0]
            emb = emb + self.unet_module.label_emb(y)

        h = x
        cache_h = None
        for id, module in enumerate(self.unet_module.input_blocks):
            transformer_options["block"] = ("input", id)
            h = forward_timestep_embed(
                module,
                h,
                emb,
                context,
                transformer_options,
                time_context=time_context,
                num_video_frames=num_video_frames,
                image_only_indicator=image_only_indicator,
            )
            h = apply_control(h, control, "input")
            if "input_block_patch" in transformer_patches:
                patch = transformer_patches["input_block_patch"]
                for p in patch:
                    h = p(h, transformer_options)

            hs.append(h)
            if "input_block_patch_after_skip" in transformer_patches:
                patch = transformer_patches["input_block_patch_after_skip"]
                for p in patch:
                    h = p(h, transformer_options)

        transformer_options["block"] = ("middle", 0)
        if self.unet_module.middle_block is not None:
            h = forward_timestep_embed(
                self.unet_module.middle_block,
                h,
                emb,
                context,
                transformer_options,
                time_context=time_context,
                num_video_frames=num_video_frames,
                image_only_indicator=image_only_indicator,
            )
        h = apply_control(h, control, "middle")

        for id, module in enumerate(self.unet_module.output_blocks):
            if id == len(self.unet_module.output_blocks) - self.cache_depth - 1:
                cache_h = h
            transformer_options["block"] = ("output", id)
            hsp = hs.pop()
            hsp = apply_control(hsp, control, "output")

            if "output_block_patch" in transformer_patches:
                patch = transformer_patches["output_block_patch"]
                for p in patch:
                    h, hsp = p(h, hsp, transformer_options)

            h = torch.cat([h, hsp], dim=1)
            del hsp
            if len(hs) > 0:
                output_shape = hs[-1].shape
            else:
                output_shape = None
            h = forward_timestep_embed(
                module,
                h,
                emb,
                context,
                transformer_options,
                output_shape,
                time_context=time_context,
                num_video_frames=num_video_frames,
                image_only_indicator=image_only_indicator,
            )
        h = h.type(x.dtype)
        if self.unet_module.predict_codebook_ids:
            return self.unet_module.id_predictor(h), cache_h
        else:
            return self.unet_module.out(h), cache_h


class FastDeepCacheUNet(Module):
    def __init__(self, unet_module, cache_layer_id, cache_block_id):
        super().__init__()
        self.unet_module = unet_module
        cache_depth = 0
        for i in range(cache_layer_id):
            cache_depth = cache_depth + self.unet_module.num_res_blocks[i]
        cache_depth = cache_depth + cache_block_id + 1
        self.cache_depth = cache_depth

    def forward(
        self,
        x,
        cache_h,
        timesteps=None,
        context=None,
        y=None,
        control=None,
        transformer_options={},
        **kwargs
    ):
        """
        Apply the model to an input batch.
        :param x: an [N x C x ...] Tensor of inputs.
        :param timesteps: a 1-D batch of timesteps.
        :param context: conditioning plugged in via crossattn
        :param y: an [N] Tensor of labels, if class-conditional.
        :return: an [N x C x ...] Tensor of outputs.
        """
        transformer_options["original_shape"] = list(x.shape)
        transformer_options["transformer_index"] = 0
        transformer_patches = transformer_options.get("patches", {})

        c_dict = kwargs.get("c", {})
        num_video_frames = c_dict.get(
            "num_video_frames", self.unet_module.default_num_video_frames
        )

        default_image_only_indicator = getattr(
            self.unet_module, "default_image_only_indicator", None
        )
        image_only_indicator = c_dict.get(
            "image_only_indicator", default_image_only_indicator
        )

        time_context = kwargs.get("time_context", None)

        assert (y is not None) == (
            self.unet_module.num_classes is not None
        ), "must specify y if and only if the model is class-conditional"
        hs = []
        t_emb = timestep_embedding(
            timesteps, self.unet_module.model_channels, repeat_only=False
        ).to(x.dtype)
        emb = self.unet_module.time_embed(t_emb)

        if self.unet_module.num_classes is not None:
            assert y.shape[0] == x.shape[0]
            emb = emb + self.unet_module.label_emb(y)

        h = x
        for id, module in enumerate(self.unet_module.input_blocks):
            transformer_options["block"] = ("input", id)
            h = forward_timestep_embed(
                module,
                h,
                emb,
                context,
                transformer_options,
                time_context=time_context,
                num_video_frames=num_video_frames,
                image_only_indicator=image_only_indicator,
            )
            h = apply_control(h, control, "input")
            if "input_block_patch" in transformer_patches:
                patch = transformer_patches["input_block_patch"]
                for p in patch:
                    h = p(h, transformer_options)

            hs.append(h)
            if "input_block_patch_after_skip" in transformer_patches:
                patch = transformer_patches["input_block_patch_after_skip"]
                for p in patch:
                    h = p(h, transformer_options)
            if id == self.cache_depth:
                break

        for id, module in enumerate(self.unet_module.output_blocks):
            if id < len(self.unet_module.output_blocks) - self.cache_depth - 1:
                continue

            if id == len(self.unet_module.output_blocks) - self.cache_depth - 1:
                h = cache_h

            transformer_options["block"] = ("output", id)
            hsp = hs.pop()
            hsp = apply_control(hsp, control, "output")

            if "output_block_patch" in transformer_patches:
                patch = transformer_patches["output_block_patch"]
                for p in patch:
                    h, hsp = p(h, hsp, transformer_options)

            h = torch.cat([h, hsp], dim=1)
            del hsp
            if len(hs) > 0:
                output_shape = hs[-1].shape
            else:
                output_shape = None
            h = forward_timestep_embed(
                module,
                h,
                emb,
                context,
                transformer_options,
                output_shape,
                time_context=time_context,
                num_video_frames=num_video_frames,
                image_only_indicator=image_only_indicator,
            )
        h = h.type(x.dtype)
        if self.unet_module.predict_codebook_ids:
            return self.unet_module.id_predictor(h), cache_h
        else:
            return self.unet_module.out(h), cache_h



"""
Install:
    pip install pytest
Uasge:
    python -m pytest test_torch2of_demo.py
"""
import torch
import oneflow as flow  # usort: skip
import unittest

import numpy as np
from onediff.infer_compiler import oneflow_compile
from onediff.infer_compiler.backends.oneflow.transform import transform_mgr


class PyTorchModel(torch.nn.Module):
    """used torch2of conversion.

    For PyTorch models, input model must inherit torch.nn.Module to utilize trace and conversion of layers.
    """

    def __init__(self):
        super().__init__()
        self.linear = torch.nn.Linear(4, 4)

    def forward(self, x):
        return self.linear(x)

    def apply_model(self, x):
        return self.forward(x)


class OneFlowModel(flow.nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = flow.nn.Linear(4, 4)

    def forward(self, x):
        return self.linear(x)

    def apply_model(self, x):
        return self.forward(x)


class TestTorch2ofDemo(unittest.TestCase):
    def judge_tensor_func(self, y_pt, y_of):
        assert type(y_pt) == type(y_of)
        assert y_pt.device == y_of.device
        y_pt = y_pt.cpu().detach().numpy()
        y_of = y_of.cpu().detach().numpy()
        assert np.allclose(y_pt, y_of, atol=1e-3, rtol=1e-3)

    def test_torch2of_demo(self):
        # Register PyTorch model to OneDiff
        cls_key = transform_mgr.get_transformed_entity_name(PyTorchModel)
        transform_mgr.update_class_proxies({cls_key: OneFlowModel})

        # Compile PyTorch model to OneFlow
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        pytorch_model = PyTorchModel().to(device)
        of_model = oneflow_compile(pytorch_model)

        # Verify conversion
        x = torch.randn(4, 4).to(device)

        #### 1. Use apply_model method
        y_pt = pytorch_model.apply_model(x)
        y_of = of_model.apply_model(x)
        self.judge_tensor_func(y_pt, y_of)

        #### 2. Use __call__ method
        y_pt = pytorch_model(x)
        y_of = of_model(x)
        self.judge_tensor_func(y_pt, y_of)


if __name__ == "__main__":
    unittest.main()



# NOTE:
# ValueError: Cannot load /home/fenrir/.cache/huggingface/hub/models--stabilityai--stable-cascade/snapshots/f2a84281d6f8db3c757195dd0c9a38dbdea90bb4/decoder because embedding.1.weight expected shape tensor(..., device='meta', size=(320, 64, 1, 1)), but got torch.Size([320, 16, 1, 1]). If you want to instead overwrite randomly initialized weights, please make sure to pass both low_cpu_mem_usage=False and ignore_mismatched_sizes=True. For more information, see also: https://github.com/huggingface/diffusers/issues/1619#issuecomment-1345604389 as an example.
# Solution is here: https://huggingface.co/stabilityai/stable-cascade/discussions/17
PRIOR_MODEL = "stabilityai/stable-cascade-prior"
PRIOR_VARIANT = None
PRIOR_CUSTOM_PIPELINE = None
PRIOR_SCHEDULER = None
PRIOR_LORA = None
PRIOR_CONTROLNET = None
PRIOR_DTYPE = "bfloat16"

DECODER_MODEL = "stabilityai/stable-cascade"
DECODER_VARIANT = None
DECODER_CUSTOM_PIPELINE = None
DECODER_SCHEDULER = None
DECODER_LORA = None
DECODER_CONTROLNET = None
DECODER_DTYPE = "float16"

PROMPT = "Anthropomorphic cat dressed as a pilot"
NEGATIVE_PROMPT = None

PRIOR_STEPS = 20
PRIOR_SEED = None
PRIOR_EXTRA_CALL_KWARGS = """{
    "guidance_scale": 4.0
}"""

DECODER_STEPS = 10
DECODER_SEED = None
DECODER_EXTRA_CALL_KWARGS = """{
    "guidance_scale": 0.0,
    "output_type": "pil"
}"""

WARMUPS = 3
BATCH = 1
HEIGHT = 1024
WIDTH = 1024
INPUT_IMAGE = None
PRIOR_CONTROL_IMAGE = None
DECODER_CONTROL_IMAGE = None
OUTPUT_IMAGE = None

import argparse
import importlib
import inspect
import json
import os
import time
from contextlib import nullcontext

import torch
from diffusers.utils import load_image
from PIL import Image, ImageDraw

import oneflow as flow  # usort: skip
from onediffx import compile_pipe


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--prior-model", type=str, default=PRIOR_MODEL)
    parser.add_argument("--prior-variant", type=str, default=PRIOR_VARIANT)
    parser.add_argument(
        "--prior-custom-pipeline", type=str, default=PRIOR_CUSTOM_PIPELINE
    )
    parser.add_argument("--prior-scheduler", type=str, default=PRIOR_SCHEDULER)
    parser.add_argument("--prior-lora", type=str, default=PRIOR_LORA)
    parser.add_argument("--prior-controlnet", type=str, default=PRIOR_CONTROLNET)
    parser.add_argument("--prior-dtype", type=str, default=PRIOR_DTYPE)

    parser.add_argument("--decoder-model", type=str, default=DECODER_MODEL)
    parser.add_argument("--decoder-variant", type=str, default=DECODER_VARIANT)
    parser.add_argument(
        "--decoder-custom-pipeline", type=str, default=DECODER_CUSTOM_PIPELINE
    )
    parser.add_argument("--decoder-scheduler", type=str, default=DECODER_SCHEDULER)
    parser.add_argument("--decoder-lora", type=str, default=DECODER_LORA)
    parser.add_argument("--decoder-controlnet", type=str, default=DECODER_CONTROLNET)
    parser.add_argument("--decoder-dtype", type=str, default=DECODER_DTYPE)

    parser.add_argument("--prompt", type=str, default=PROMPT)
    parser.add_argument("--negative-prompt", type=str, default=NEGATIVE_PROMPT)

    parser.add_argument("--prior-steps", type=int, default=PRIOR_STEPS)
    parser.add_argument("--prior-seed", type=int, default=PRIOR_SEED)
    parser.add_argument(
        "--prior-extra-call-kwargs", type=str, default=PRIOR_EXTRA_CALL_KWARGS
    )

    parser.add_argument("--decoder-steps", type=int, default=DECODER_STEPS)
    parser.add_argument("--decoder-seed", type=int, default=DECODER_SEED)
    parser.add_argument(
        "--decoder-extra-call-kwargs", type=str, default=DECODER_EXTRA_CALL_KWARGS
    )

    parser.add_argument("--warmups", type=int, default=WARMUPS)
    parser.add_argument("--batch", type=int, default=BATCH)
    parser.add_argument("--height", type=int, default=HEIGHT)
    parser.add_argument("--width", type=int, default=WIDTH)
    parser.add_argument("--input-image", type=str, default=INPUT_IMAGE)
    parser.add_argument("--prior-control-image", type=str, default=PRIOR_CONTROL_IMAGE)
    parser.add_argument(
        "--decoder-control-image", type=str, default=DECODER_CONTROL_IMAGE
    )
    parser.add_argument("--output-image", type=str, default=OUTPUT_IMAGE)
    parser.add_argument(
        "--compiler",
        type=str,
        default="oneflow",
        choices=["none", "oneflow", "compile", "compile-max-autotune"],
    )
    return parser.parse_args()


def load_pipe(
    pipeline_cls,
    model_name,
    variant=None,
    custom_pipeline=None,
    scheduler=None,
    lora=None,
    controlnet=None,
    dtype=torch.float16,
):
    extra_kwargs = {}
    if custom_pipeline is not None:
        extra_kwargs["custom_pipeline"] = custom_pipeline
    if variant is not None:
        extra_kwargs["variant"] = variant
    if controlnet is not None:
        from diffusers import ControlNetModel

        controlnet = ControlNetModel.from_pretrained(
            controlnet,
            torch_dtype=dtype,
        )
        extra_kwargs["controlnet"] = controlnet
    if os.path.exists(os.path.join(model_name, "calibrate_info.txt")):
        from onediff.quantization import QuantPipeline

        pipe = QuantPipeline.from_pretrained(
            pipeline_cls, model_name, torch_dtype=dtype, **extra_kwargs
        )
    else:
        pipe = pipeline_cls.from_pretrained(
            model_name, torch_dtype=dtype, **extra_kwargs
        )
    if scheduler is not None:
        scheduler_cls = getattr(importlib.import_module("diffusers"), scheduler)
        pipe.scheduler = scheduler_cls.from_config(pipe.scheduler.config)
    if lora is not None:
        pipe.load_lora_weights(lora)
        pipe.fuse_lora()
    pipe.safety_checker = None
    pipe.to(torch.device("cuda"))
    return pipe


class IterationProfiler:
    def __init__(self):
        self.begin = None
        self.end = None
        self.num_iterations = 0

    def get_iter_per_sec(self):
        if self.begin is None or self.end is None:
            return None
        self.end.synchronize()
        dur = self.begin.elapsed_time(self.end)
        return self.num_iterations / dur * 1000.0

    def callback_on_step_end(self, pipe, i, t, callback_kwargs={}):
        if self.begin is None:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.begin = event
        else:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.end = event
            self.num_iterations += 1
        return callback_kwargs


def main():
    args = parse_args()
    assert args.input_image is None
    try:
        from diffusers import StableCascadeDecoderPipeline, StableCascadePriorPipeline
    except ImportError:
        raise ImportError(
            "Please install `diffusers` from this branch while the PR is WIP: `pip install --upgrade git+https://github.com/kashif/diffusers.git@a3dc21385b7386beb3dab3a9845962ede6765887`"
        )

    prior_pipe = load_pipe(
        StableCascadePriorPipeline,
        args.prior_model,
        variant=args.prior_variant,
        custom_pipeline=args.prior_custom_pipeline,
        scheduler=args.prior_scheduler,
        lora=args.prior_lora,
        controlnet=args.prior_controlnet,
        dtype=getattr(torch, args.prior_dtype),
    )

    patch_oneflow_prior_fp16_overflow = nullcontext
    if prior_pipe.dtype == torch.float16:
        if args.compiler == "oneflow":
            from patch_stable_cascade_of import patch_oneflow_prior_fp16_overflow
        # Dynamic patching would fail with oneflow
        from patch_stable_cascade import patch_prior_fp16_overflow

        prior_pipe.prior = patch_prior_fp16_overflow(prior_pipe.prior)

    decoder_pipe = load_pipe(
        StableCascadeDecoderPipeline,
        args.decoder_model,
        variant=args.decoder_variant,
        custom_pipeline=args.decoder_custom_pipeline,
        scheduler=args.decoder_scheduler,
        lora=args.decoder_lora,
        controlnet=args.decoder_controlnet,
        dtype=getattr(torch, args.decoder_dtype),
    )

    height = args.height
    width = args.width

    if args.compiler == "none":
        pass
    elif args.compiler == "oneflow":
        prior_pipe = compile_pipe(prior_pipe)
        decoder_pipe = compile_pipe(decoder_pipe)
    elif args.compiler in ("compile", "compile-max-autotune"):
        from patch_stable_cascade import patch_torch_compile

        patch_torch_compile()

        mode = "max-autotune" if args.compiler == "compile-max-autotune" else None

        prior_pipe.prior = torch.compile(prior_pipe.prior, mode=mode)
        if hasattr(prior_pipe, "controlnet"):
            prior_pipe.controlnet = torch.compile(prior_pipe.controlnet, mode=mode)

        decoder_pipe.decoder = torch.compile(decoder_pipe.decoder, mode=mode)
        if hasattr(decoder_pipe, "controlnet"):
            decoder_pipe.controlnet = torch.compile(decoder_pipe.controlnet, mode=mode)
    else:
        raise ValueError(f"Unknown compiler: {args.compiler}")

    if args.input_image is None:
        input_image = None
    else:
        input_image = load_image(args.input_image)
        input_image = input_image.resize((width, height), Image.LANCZOS)

    if args.prior_control_image is None:
        if args.prior_controlnet is None:
            prior_control_image = None
        else:
            prior_control_image = Image.new("RGB", (width, height))
            draw = ImageDraw.Draw(prior_control_image)
            draw.ellipse(
                (args.width // 4, height // 4, args.width // 4 * 3, height // 4 * 3),
                fill=(255, 255, 255),
            )
            del draw
    else:
        prior_control_image = load_image(args.prior_control_image)
        prior_control_image = prior_control_image.resize((width, height), Image.LANCZOS)

    if args.decoder_control_image is None:
        if args.decoder_controlnet is None:
            decoder_control_image = None
        else:
            decoder_control_image = Image.new("RGB", (width, height))
            draw = ImageDraw.Draw(decoder_control_image)
            draw.ellipse(
                (args.width // 4, height // 4, args.width // 4 * 3, height // 4 * 3),
                fill=(255, 255, 255),
            )
            del draw
    else:
        decoder_control_image = load_image(args.decoder_control_image)
        decoder_control_image = decoder_control_image.resize(
            (width, height), Image.LANCZOS
        )

    def get_prior_kwarg_inputs():
        kwarg_inputs = dict(
            prompt=args.prompt,
            negative_prompt=args.negative_prompt,
            height=height,
            width=width,
            num_inference_steps=args.prior_steps,
            num_images_per_prompt=args.batch,
            generator=None
            if args.prior_seed is None
            else torch.Generator(device="cuda").manual_seed(args.prior_seed),
            **(
                dict()
                if args.prior_extra_call_kwargs is None
                else json.loads(args.prior_extra_call_kwargs)
            ),
        )
        if input_image is not None:
            kwarg_inputs["image"] = input_image
        if prior_control_image is not None:
            if input_image is None:
                kwarg_inputs["image"] = prior_control_image
            else:
                kwarg_inputs["control_image"] = prior_control_image
        return kwarg_inputs

    def get_decoder_kwarg_inputs():
        kwarg_inputs = dict(
            prompt=args.prompt,
            negative_prompt=args.negative_prompt,
            num_inference_steps=args.decoder_steps,
            num_images_per_prompt=args.batch,
            generator=None
            if args.decoder_seed is None
            else torch.Generator(device="cuda").manual_seed(args.decoder_seed),
            **(
                dict()
                if args.decoder_extra_call_kwargs is None
                else json.loads(args.decoder_extra_call_kwargs)
            ),
        )
        if input_image is not None:
            kwarg_inputs["image"] = input_image
        if decoder_control_image is not None:
            if input_image is None:
                kwarg_inputs["image"] = decoder_control_image
            else:
                kwarg_inputs["control_image"] = decoder_control_image
        return kwarg_inputs

    # NOTE: Warm it up.
    # The initial calls will trigger compilation and might be very slow.
    # After that, it should be very fast.
    if args.warmups > 0:
        print("Begin warmup")
        for _ in range(args.warmups):
            with patch_oneflow_prior_fp16_overflow():
                prior_output = prior_pipe(**get_prior_kwarg_inputs())
            decoder_pipe(
                image_embeddings=prior_output.image_embeddings.to(
                    dtype=getattr(torch, args.decoder_dtype)
                ),
                **get_decoder_kwarg_inputs(),
            )
        print("End warmup")

    # Let"s see it!
    # Note: Progress bar might work incorrectly due to the async nature of CUDA.
    prior_kwarg_inputs = get_prior_kwarg_inputs()
    prior_iter_profiler = IterationProfiler()
    if "callback_on_step_end" in inspect.signature(prior_pipe).parameters:
        prior_kwarg_inputs[
            "callback_on_step_end"
        ] = prior_iter_profiler.callback_on_step_end
    elif "callback" in inspect.signature(prior_pipe).parameters:
        prior_kwarg_inputs["callback"] = prior_iter_profiler.callback_on_step_end

    decoder_kwarg_inputs = get_decoder_kwarg_inputs()
    decoder_iter_profiler = IterationProfiler()
    if "callback_on_step_end" in inspect.signature(decoder_pipe).parameters:
        decoder_kwarg_inputs[
            "callback_on_step_end"
        ] = decoder_iter_profiler.callback_on_step_end
    elif "callback" in inspect.signature(decoder_pipe).parameters:
        decoder_kwarg_inputs["callback"] = decoder_iter_profiler.callback_on_step_end

    prior_begin = time.time()
    with patch_oneflow_prior_fp16_overflow():
        prior_output = prior_pipe(**prior_kwarg_inputs)
    prior_end = time.time()

    decoder_begin = time.time()
    output_images = decoder_pipe(
        image_embeddings=prior_output.image_embeddings.to(
            dtype=getattr(torch, args.decoder_dtype)
        ),
        **decoder_kwarg_inputs,
    ).images
    decoder_end = time.time()

    print("=======================================")

    print(f"Prior Inference time: {prior_end - prior_begin:.3f}s")
    prior_iter_per_sec = prior_iter_profiler.get_iter_per_sec()
    if prior_iter_per_sec is not None:
        print(f"Prior Iterations per second: {prior_iter_per_sec:.3f}")

    print(f"Decoder Inference time: {decoder_end - decoder_begin:.3f}s")
    decoder_iter_per_sec = decoder_iter_profiler.get_iter_per_sec()
    if decoder_iter_per_sec is not None:
        print(f"Decoder Iterations per second: {decoder_iter_per_sec:.3f}")

    cuda_mem_after_used = flow._oneflow_internal.GetCUDAMemoryUsed()
    host_mem_after_used = flow._oneflow_internal.GetCPUMemoryUsed()
    print(f"CUDA Mem after: {cuda_mem_after_used / 1024:.3f}GiB")
    print(f"Host Mem after: {host_mem_after_used / 1024:.3f}GiB")
    print("=======================================")

    if args.output_image is not None:
        output_images[0].save(args.output_image)
    else:
        print("Please set `--output-image` to save the output image")


if __name__ == "__main__":
    main()



import inspect

import comfy
from comfy.controlnet import ControlLora, ControlLoraOps, ControlNet


class OneDiffControlLora(ControlLora):
    @classmethod
    def from_controllora(cls, controlnet: ControlLora, *, compile_fn: callable = None):
        init_parameters = set(inspect.signature(cls.__init__).parameters.keys())
        init_dict = {
            attr: getattr(controlnet, attr)
            for attr in init_parameters
            if attr != "self"
        }
        c = cls(**init_dict)
        controlnet.copy_to(c)
        c._control_model = None
        c._compile_fn = compile_fn
        return c

    def pre_run(self, model, percent_to_timestep_function):
        ControlNet.pre_run(self, model, percent_to_timestep_function)

        self.manual_cast_dtype = model.manual_cast_dtype
        dtype = model.get_dtype()
        if self.manual_cast_dtype is None:

            class control_lora_ops(ControlLoraOps, comfy.ops.disable_weight_init):
                pass

        else:

            class control_lora_ops(ControlLoraOps, comfy.ops.manual_cast):
                pass

            dtype = self.manual_cast_dtype
        if self._control_model is None:
            controlnet_config = model.model_config.unet_config.copy()
            controlnet_config.pop("out_channels")
            controlnet_config["hint_channels"] = self.control_weights[
                "input_hint_block.0.weight"
            ].shape[1]
            controlnet_config["operations"] = control_lora_ops
            controlnet_config["dtype"] = dtype
            self.control_model = comfy.cldm.cldm.ControlNet(**controlnet_config)
            self.control_model.to(comfy.model_management.get_torch_device())
            self._control_model = self._compile_fn(self.control_model)

        self.control_model = self._control_model
        diffusion_model = model.diffusion_model
        sd = diffusion_model.state_dict()
        # cm = self.control_model.state_dict()

        for k in sd:
            weight = sd[k]
            try:
                comfy.utils.set_attr_param(self.control_model, k, weight)
            except:
                pass

        for k in self.control_weights:
            if k not in {"lora_controlnet"}:
                comfy.utils.set_attr_param(
                    self.control_model,
                    k,
                    self.control_weights[k]
                    .to(dtype)
                    .to(comfy.model_management.get_torch_device()),
                )

    def cleanup(self):
        pass

    def copy(self):
        init_parameters = set(inspect.signature(type(self).__init__).parameters.keys())
        init_dict = {
            attr: getattr(self, attr) for attr in init_parameters if attr != "self"
        }
        c = type(self)(**init_dict)

        self.copy_to(c)
        c._control_model = self._control_model
        c._compile_fn = self._compile_fn
        return c



"""hijack ComfyUI/custom_nodes/ComfyUI_IPAdapter_plus/IPAdapterPlus.py

hijack ComfyUI/custom_nodes/ComfyUI_IPAdapter_plus/IPAdapterPlus.py"""

import functools

from onediff.utils.log_utils import logger

from ..utils.booster_utils import is_using_oneflow_backend
from ._config import ipadapter_plus_hijacker, ipadapter_plus_pt
from .set_model_patch_replace import set_model_patch_replace_v2

set_model_patch_replace_fn_pt = ipadapter_plus_pt.IPAdapterPlus.set_model_patch_replace
ipadapter_execute = ipadapter_plus_pt.IPAdapterPlus.ipadapter_execute


def cache_init(original_init):
    @functools.wraps(original_init)
    def new_init(self, ipadapter_model, *args, **kwargs):
        cache_key = args + tuple(kwargs.values())
        # import time
        # t0 = time.time()
        if cache_key in ipadapter_model:
            cached_instance = ipadapter_model[cache_key]
            self.__dict__ = cached_instance.__dict__
            logger.debug("Using cached IPAdapter instance")
        else:
            original_init(self, ipadapter_model, *args, **kwargs)
            ipadapter_model[cache_key] = self
            logger.debug("Caching new IPAdapter instance")

        # print(f'dur: {time.time() - t0}')

    return new_init


def cond_func(org_fn, model, *args, **kwargs):
    return is_using_oneflow_backend(model)


def ipadapter_execute_of(org_fn, model, *args, **kwargs):
    IPAdapter = ipadapter_plus_pt.IPAdapterPlus.IPAdapter
    _original_init = IPAdapter.__init__
    IPAdapter.__init__ = cache_init(_original_init)
    output = org_fn(model, *args, **kwargs)
    IPAdapter.__init__ = _original_init
    return output


ipadapter_plus_hijacker.register(
    set_model_patch_replace_fn_pt, set_model_patch_replace_v2, cond_func
)
ipadapter_plus_hijacker.register(ipadapter_execute, ipadapter_execute_of, cond_func)



"""align attention with ipadapter
https://github.com/cubiq/ComfyUI_InstantID/blob/main/CrossAttentionPatch.py
https://github.com/cubiq/ComfyUI_IPAdapter_plus/blob/main/CrossAttentionPatch.py
"""

import math

import torch
import torch.nn.functional as F
from comfy.ldm.modules.attention import attention_pytorch as optimized_attention


def tensor_to_size(source, dest_size):
    if source.dim() == 0:
        print("x is a scalar (no dimensions)")
        return source
    if isinstance(dest_size, torch.Tensor):
        dest_size = dest_size.shape[0]
    source_size = source.shape[0]

    if source_size < dest_size:
        shape = [dest_size - source_size] + [1] * (source.dim() - 1)
        source = torch.cat((source, source[-1:].repeat(shape)), dim=0)
    elif source_size > dest_size:
        source = source[:dest_size]
    return source


def get_weight_subidxs(weight, ad_params, sub_idxs):
    return weight[ad_params[sub_idxs]]


class Attn2Replace:
    def __init__(self, callback=None, **kwargs):
        self.callback = [callback]
        self.kwargs = [kwargs]

        self.forward_patch_key = id(self)
        self._use_crossAttention_patch = True
        self.cache_map = {}  # {ui_index, index}
        self._bind_model = None
        self.optimized_attention = optimized_attention

    def get_bind_model(self):
        return self._bind_model

    def add(self, callback, **kwargs):
        self.callback.append(callback)
        self.kwargs.append(kwargs)

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __call__(self, q, k, v, extra_options):
        dtype = q.dtype
        out = self.optimized_attention(q, k, v, extra_options["n_heads"])
        sigma = extra_options["sigmas"] if "sigmas" in extra_options else 999999999.9

        patch_kwargs = extra_options["_attn2"].get(self.forward_patch_key)
        for i, callback in enumerate(self.callback):
            if (
                sigma <= self.kwargs[i]["sigma_start"]
                and sigma >= self.kwargs[i]["sigma_end"]
            ):
                inputs = self.kwargs[i]
                inputs.update(**patch_kwargs[i])
                out = out + callback(
                    out,
                    q,
                    k,
                    v,
                    extra_options,
                    optimized_attention=self.optimized_attention,
                    **inputs,
                )

        return out.to(dtype=dtype)

    def __deepcopy__(self, memo):
        # print("Warning: CrossAttentionPatch is not deepcopiable.", '-'*20)
        return self


def ipadapter_attention(
    out,
    q,
    k,
    v,
    extra_options,
    module_key="",
    ipadapter=None,
    weight=1.0,
    cond=None,
    cond_alt=None,
    uncond=None,
    weight_type="linear",
    mask=None,
    sigma_start=0.0,
    sigma_end=1.0,
    unfold_batch=False,
    embeds_scaling="V only",
    optimized_attention=None,
    **kwargs
):
    dtype = q.dtype
    cond_or_uncond = extra_options["cond_or_uncond"]
    block_type = extra_options["block"][0]
    # block_id = extra_options["block"][1]
    t_idx = extra_options["transformer_index"]
    layers = 11 if "101_to_k_ip" in ipadapter.ip_layers.to_kvs else 16
    k_key = module_key + "_to_k_ip"
    v_key = module_key + "_to_v_ip"

    # extra options for AnimateDiff
    ad_params = extra_options["ad_params"] if "ad_params" in extra_options else None

    b = q.shape[0]
    seq_len = q.shape[1]
    batch_prompt = b // len(cond_or_uncond)
    _, _, oh, ow = extra_options["original_shape"]

    if weight_type == "ease in":
        weight = weight * (0.05 + 0.95 * (1 - t_idx / layers))
    elif weight_type == "ease out":
        weight = weight * (0.05 + 0.95 * (t_idx / layers))
    elif weight_type == "ease in-out":
        weight = weight * (0.05 + 0.95 * (1 - abs(t_idx - (layers / 2)) / (layers / 2)))
    elif weight_type == "reverse in-out":
        weight = weight * (0.05 + 0.95 * (abs(t_idx - (layers / 2)) / (layers / 2)))
    elif weight_type == "weak input" and block_type == "input":
        weight = weight * 0.2
    elif weight_type == "weak middle" and block_type == "middle":
        weight = weight * 0.2
    elif weight_type == "weak output" and block_type == "output":
        weight = weight * 0.2
    elif weight_type == "strong middle" and (
        block_type == "input" or block_type == "output"
    ):
        weight = weight * 0.2
    elif isinstance(weight, dict):
        if t_idx not in weight:
            return 0

        if weight_type == "style transfer precise":
            if layers == 11 and t_idx == 3:
                uncond = cond
                cond = cond * 0
            elif layers == 16 and (t_idx == 4 or t_idx == 5):
                uncond = cond
                cond = cond * 0

        elif weight_type == "composition precise":
            if layers == 11 and t_idx != 3:
                uncond = cond
                cond = cond * 0
            elif layers == 16 and (t_idx != 4 and t_idx != 5):
                uncond = cond
                cond = cond * 0

        weight = weight[t_idx]

        if cond_alt is not None and t_idx in cond_alt:
            cond = cond_alt[t_idx]
            del cond_alt

    if unfold_batch:
        # Check AnimateDiff context window
        if ad_params is not None and ad_params["sub_idxs"] is not None:
            if isinstance(weight, torch.Tensor) and weight.dim() != 0:
                weight = tensor_to_size(weight, ad_params["full_length"])
                weight = get_weight_subidxs(weight, ad_params, "sub_idxs")
                # if torch.all(weight == 0):
                #     return 0
                weight = weight.repeat(
                    len(cond_or_uncond), 1, 1
                )  # repeat for cond and uncond
            # elif weight == 0:
            #     return 0

            # if image length matches or exceeds full_length get sub_idx images
            if cond.shape[0] >= ad_params["full_length"]:
                cond = get_weight_subidxs(cond, ad_params, "sub_idxs")
                uncond = get_weight_subidxs(uncond, ad_params, "sub_idxs")
            # otherwise get sub_idxs images
            else:
                cond = tensor_to_size(cond, ad_params["full_length"])
                uncond = tensor_to_size(uncond, ad_params["full_length"])
                cond = cond[ad_params["sub_idxs"]]
                uncond = uncond[ad_params["sub_idxs"]]
        else:
            if isinstance(weight, torch.Tensor) and weight.dim() != 0:
                weight = tensor_to_size(weight, batch_prompt)
                # if torch.all(weight == 0):
                #     return 0
                weight = weight.repeat(
                    len(cond_or_uncond), 1, 1
                )  # repeat for cond and uncond
            # elif weight == 0:
            #     return 0

            cond = tensor_to_size(cond, batch_prompt)
            uncond = tensor_to_size(uncond, batch_prompt)

        k_cond = ipadapter.ip_layers.to_kvs[k_key](cond)
        k_uncond = ipadapter.ip_layers.to_kvs[k_key](uncond)
        v_cond = ipadapter.ip_layers.to_kvs[v_key](cond)
        v_uncond = ipadapter.ip_layers.to_kvs[v_key](uncond)
    else:
        # TODO: should we always convert the weights to a tensor?
        if isinstance(weight, torch.Tensor) and weight.dim() != 0:
            weight = tensor_to_size(weight, batch_prompt)
            # if torch.all(weight == 0):
            #     return 0
            weight = weight.repeat(
                len(cond_or_uncond), 1, 1
            )  # repeat for cond and uncond
        # elif weight == 0:
        #     return 0

        k_cond = ipadapter.ip_layers.to_kvs[k_key](cond).repeat(batch_prompt, 1, 1)
        k_uncond = ipadapter.ip_layers.to_kvs[k_key](uncond).repeat(batch_prompt, 1, 1)
        v_cond = ipadapter.ip_layers.to_kvs[v_key](cond).repeat(batch_prompt, 1, 1)
        v_uncond = ipadapter.ip_layers.to_kvs[v_key](uncond).repeat(batch_prompt, 1, 1)

    if len(cond_or_uncond) == 3:  # TODO: conxl, I need to check this
        ip_k = torch.cat([(k_cond, k_uncond, k_cond)[i] for i in cond_or_uncond], dim=0)
        ip_v = torch.cat([(v_cond, v_uncond, v_cond)[i] for i in cond_or_uncond], dim=0)
    else:
        ip_k = torch.cat([(k_cond, k_uncond)[i] for i in cond_or_uncond], dim=0)
        ip_v = torch.cat([(v_cond, v_uncond)[i] for i in cond_or_uncond], dim=0)

    if embeds_scaling == "K+mean(V) w/ C penalty":
        scaling = float(ip_k.shape[2]) / 1280.0
        weight = weight * scaling
        ip_k = ip_k * weight
        ip_v_mean = torch.mean(ip_v, dim=1, keepdim=True)
        ip_v = (ip_v - ip_v_mean) + ip_v_mean * weight
        out_ip = optimized_attention(q, ip_k, ip_v, extra_options["n_heads"])
        del ip_v_mean
    elif embeds_scaling == "K+V w/ C penalty":
        scaling = float(ip_k.shape[2]) / 1280.0
        weight = weight * scaling
        ip_k = ip_k * weight
        ip_v = ip_v * weight
        out_ip = optimized_attention(q, ip_k, ip_v, extra_options["n_heads"])
    elif embeds_scaling == "K+V":
        ip_k = ip_k * weight
        ip_v = ip_v * weight
        out_ip = optimized_attention(q, ip_k, ip_v, extra_options["n_heads"])
    else:
        # ip_v = ip_v * weight
        out_ip = optimized_attention(q, ip_k, ip_v, extra_options["n_heads"])
        out_ip = out_ip * weight  # I'm doing this to get the same results as before

    if mask is not None:
        mask_h = oh / math.sqrt(oh * ow / seq_len)
        mask_h = int(mask_h) + int((seq_len % int(mask_h)) != 0)
        mask_w = seq_len // mask_h

        # check if using AnimateDiff and sliding context window
        if (
            mask.shape[0] > 1
            and ad_params is not None
            and ad_params["sub_idxs"] is not None
        ):
            # if mask length matches or exceeds full_length, get sub_idx masks
            if mask.shape[0] >= ad_params["full_length"]:
                mask = torch.Tensor(mask[ad_params["sub_idxs"]])
                mask = F.interpolate(
                    mask.unsqueeze(1), size=(mask_h, mask_w), mode="bilinear"
                ).squeeze(1)
            else:
                mask = F.interpolate(
                    mask.unsqueeze(1), size=(mask_h, mask_w), mode="bilinear"
                ).squeeze(1)
                mask = tensor_to_size(mask, ad_params["full_length"])
                mask = mask[ad_params["sub_idxs"]]
        else:
            mask = F.interpolate(
                mask.unsqueeze(1), size=(mask_h, mask_w), mode="bilinear"
            ).squeeze(1)
            mask = tensor_to_size(mask, batch_prompt)

        mask = mask.repeat(len(cond_or_uncond), 1, 1)
        mask = mask.view(mask.shape[0], -1, 1).repeat(1, 1, out.shape[2])

        # covers cases where extreme aspect ratios can cause the mask to have a wrong size
        mask_len = mask_h * mask_w
        if mask_len < seq_len:
            pad_len = seq_len - mask_len
            pad1 = pad_len // 2
            pad2 = pad_len - pad1
            mask = F.pad(mask, (0, 0, pad1, pad2), value=0.0)
        elif mask_len > seq_len:
            crop_start = (mask_len - seq_len) // 2
            mask = mask[:, crop_start : crop_start + seq_len, :]

        out_ip = out_ip * mask

    # out = out + out_ip

    return out_ip.to(dtype=dtype)


def is_crossAttention_patch(module) -> bool:
    return getattr(module, "_use_crossAttention_patch", False)


def pulid_attention(
    out,
    q,
    k,
    v,
    extra_options,
    module_key="",
    pulid=None,
    cond=None,
    uncond=None,
    weight=1.0,
    ortho=False,
    ortho_v2=False,
    mask=None,
    optimized_attention=None,
    **kwargs
):
    k_key = module_key + "_to_k_ip"
    v_key = module_key + "_to_v_ip"

    dtype = q.dtype
    seq_len = q.shape[1]
    cond_or_uncond = extra_options["cond_or_uncond"]
    b = q.shape[0]
    batch_prompt = b // len(cond_or_uncond)
    _, _, oh, ow = extra_options["original_shape"]

    # conds = torch.cat([uncond.repeat(batch_prompt, 1, 1), cond.repeat(batch_prompt, 1, 1)], dim=0)
    # zero_tensor = torch.zeros((conds.size(0), num_zero, conds.size(-1)), dtype=conds.dtype, device=conds.device)
    # conds = torch.cat([conds, zero_tensor], dim=1)
    # ip_k = pulid.ip_layers.to_kvs[k_key](conds)
    # ip_v = pulid.ip_layers.to_kvs[v_key](conds)

    k_cond = pulid.ip_layers.to_kvs[k_key](cond).repeat(batch_prompt, 1, 1)
    k_uncond = pulid.ip_layers.to_kvs[k_key](uncond).repeat(batch_prompt, 1, 1)
    v_cond = pulid.ip_layers.to_kvs[v_key](cond).repeat(batch_prompt, 1, 1)
    v_uncond = pulid.ip_layers.to_kvs[v_key](uncond).repeat(batch_prompt, 1, 1)
    ip_k = torch.cat([(k_cond, k_uncond)[i] for i in cond_or_uncond], dim=0)
    ip_v = torch.cat([(v_cond, v_uncond)[i] for i in cond_or_uncond], dim=0)

    out_ip = optimized_attention(q, ip_k, ip_v, extra_options["n_heads"])

    if ortho:
        out = out.to(dtype=torch.float32)
        out_ip = out_ip.to(dtype=torch.float32)
        projection = (
            torch.sum((out * out_ip), dim=-2, keepdim=True)
            / torch.sum((out * out), dim=-2, keepdim=True)
            * out
        )
        orthogonal = out_ip - projection
        out_ip = weight * orthogonal
    elif ortho_v2:
        out = out.to(dtype=torch.float32)
        out_ip = out_ip.to(dtype=torch.float32)
        attn_map = q @ ip_k.transpose(-2, -1)
        attn_mean = attn_map.softmax(dim=-1).mean(dim=1, keepdim=True)
        attn_mean = attn_mean[:, :, :5].sum(dim=-1, keepdim=True)
        projection = (
            torch.sum((out * out_ip), dim=-2, keepdim=True)
            / torch.sum((out * out), dim=-2, keepdim=True)
            * out
        )
        orthogonal = out_ip + (attn_mean - 1) * projection
        out_ip = weight * orthogonal
    else:
        out_ip = out_ip * weight

    if mask is not None:
        mask_h = oh / math.sqrt(oh * ow / seq_len)
        mask_h = int(mask_h) + int((seq_len % int(mask_h)) != 0)
        mask_w = seq_len // mask_h

        mask = F.interpolate(
            mask.unsqueeze(1), size=(mask_h, mask_w), mode="bilinear"
        ).squeeze(1)
        mask = tensor_to_size(mask, batch_prompt)

        mask = mask.repeat(len(cond_or_uncond), 1, 1)
        mask = mask.view(mask.shape[0], -1, 1).repeat(1, 1, out.shape[2])

        # covers cases where extreme aspect ratios can cause the mask to have a wrong size
        mask_len = mask_h * mask_w
        if mask_len < seq_len:
            pad_len = seq_len - mask_len
            pad1 = pad_len // 2
            pad2 = pad_len - pad1
            mask = F.pad(mask, (0, 0, pad1, pad2), value=0.0)
        elif mask_len > seq_len:
            crop_start = (mask_len - seq_len) // 2
            mask = mask[:, crop_start : crop_start + seq_len, :]

        out_ip = out_ip * mask

    return out_ip.to(dtype=dtype)
